# -*- coding: utf-8 -*-


def inter_line_line(Z1, Z2, Z3, Z4):
    """INTER_LINE_LINE find the intersection between two lines defined by two
    complex

    Parameters
    ----------
    Z1 : complex
        Complex coordinate of a point on the first line
    Z2 : complex
        Complex coordinate of another point on the first line
    Z3 : complex
        Complex coordinate of a point on the second line
    Z4 : complex
        Complex coordinate of another point on the second line

    Returns
    -------
        Zlist: list
            List of the complex coordinates of the intersection
            return [Z1, Z2] if the 4 points are aligned
    """

    # Check if the points are aligned
    Z12 = Z1 - Z2
    Z13 = Z1 - Z3
    Z14 = Z1 - Z4
    if (
        Z12.real * Z13.imag - Z12.imag * Z13.real == 0
        and Z12.real * Z14.imag - Z12.imag * Z14.real == 0
    ):
        return [Z1, Z2]

    # Compute the line equation
    (A1, B1, C1) = find_line_eq(Z1, Z2)
    (A2, B2, C2) = find_line_eq(Z3, Z4)
    # Compute the intersection
    D = A1 * B2 - B1 * A2
    Dx = C1 * B2 - B1 * C2
    Dy = A1 * C2 - C1 * A2
    if D != 0:
        x = Dx / D
        y = Dy / D
        return [x + 1j * y]
    else:
        return []


def find_line_eq(Z1, Z2):
    """Find the line equation (Ax+By=C)

    Parameters
    ----------
    Z1 : complex
        Complex coordinate of a point on the line
    Z2 : complex
         Complex coordinate of another point on the line

    Returns
    -------
    A, B, C : (float, float, float)
        Line equation parameters
    """
    A = Z1.imag - Z2.imag
    B = Z2.real - Z1.real
    C = Z1.real * Z2.imag - Z2.real * Z1.imag
    return (A, B, -C)



import os
from os.path import join

from ...Functions.FEMM.draw_FEMM import draw_FEMM
from ...Functions.Electrical.coordinate_transformation import n2dq
from ...Classes._FEMMHandler import _FEMMHandler
from ...Classes.OutMagFEMM import OutMagFEMM
from numpy import linspace, pi, split
from SciDataTool.Classes.Data1D import Data1D


def comp_fluxlinkage(obj, output):
    """Compute the flux linkage using FEMM and electrical output reference currents

    Parameters
    ----------
    obj : FluxLinkFEMM or IndMagFEMM
        a FluxLinkFEMM object or an IndMagFEMM object
    output : Output
        an Output object

    Return
    ------
    fluxdq : ndarray
        the calculated fluxlinkage
    """
    # get some machine and simulation parameters
    qs = output.simu.machine.stator.winding.qs
    zp = output.simu.machine.stator.get_pole_pair_number()
    Nt_tot = obj.Nt_tot
    rot_dir = output.get_rot_dir()

    # Get save path
    str_EEC = "EEC"

    path_res = output.get_path_result()

    save_dir = join(path_res, "Femm")
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    if output.simu.machine.name not in [None, ""]:
        file_name = output.simu.machine.name + "_" + str_EEC + ".fem"
    elif output.simu.name not in [None, ""]:
        file_name = output.simu.name + "_" + str_EEC + ".fem"
    else:  # Default name
        file_name = "FEMM_" + str_EEC + ".fem"

    path_save = join(save_dir, file_name)

    # Set the symmetry factor according to the machine
    if obj.is_periodicity_a:
        (
            sym,
            is_antiper_a,
            _,
            _,
        ) = obj.parent.parent.parent.parent.get_machine_periodicity()
        if is_antiper_a:
            sym = sym * 2
    else:
        sym = 1
        is_antiper_a = False

    # store orignal elec and make a copy to do temp. modifications
    elec = output.elec
    output.elec = elec.copy()

    # Set rotor angle for the FEMM simulation
    angle_offset_initial = output.get_angle_offset_initial()
    angle_rotor = (
        linspace(0, -1 * rot_dir * 2 * pi / sym, Nt_tot, endpoint=False)
        + angle_offset_initial
    )

    # modify some quantities
    output.elec.Time = Data1D(
        name="time",
        unit="s",
        values=(angle_rotor - angle_rotor[0]) / (2 * pi * output.elec.N0 / 60),
    )
    output.elec.Is = None  # to compute Is from Id_ref and Iq_ref (that are mean val.)
    output.elec.Is = output.elec.get_Is()  # TODO get_Is disregards initial rotor angle

    # Open FEMM
    femm = _FEMMHandler()
    if output.elec.internal is None:
        output.elec.internal = OutMagFEMM()
    output.elec.internal.handler_list.append(femm)

    # Setup the FEMM simulation
    # Geometry building and assigning property in FEMM
    FEMM_dict = draw_FEMM(
        femm=femm,
        output=output,
        is_mmfr=1,
        is_mmfs=1,
        sym=sym,
        is_antiper=is_antiper_a,
        type_calc_leakage=obj.type_calc_leakage,
        kgeo_fineness=obj.Kgeo_fineness,  # TODO fix inconsistent lower/upper case
        path_save=path_save,
    )

    # Solve for all time step and store all the results in output
    Phi_wind = obj.solve_FEMM(femm, output, sym, FEMM_dict)

    # Close FEMM after simulation
    femm.closefemm()
    output.elec.internal.handler_list.remove(femm)

    # Define d axis angle for the d,q transform
    d_angle = (angle_rotor - angle_offset_initial) * zp
    fluxdq = split(n2dq(Phi_wind, d_angle, n=qs), 2, axis=1)

    # restore the original elec
    output.elec = elec

    return fluxdq



# -*- coding: utf-8 -*-

from numpy import exp


def comp_radius(self):
    """Compute the radius of the min and max circle that contains the hole

    Parameters
    ----------
    self : HoleM51
        A HoleM51 object

    Returns
    -------
    (Rmin, Rmax): tuple
        Radius of the circle that contains the hole [m]

    """

    Rext = self.get_Rext()

    Rmax = Rext - self.H1
    Rmin = Rext - self.H0 - self.H2

    return (Rmin, Rmax)



# -*- coding: utf-8 -*-

from numpy import pi


def comp_torque(self, output):
    """Compute the electrical average torque

    Parameters
    ----------
    self : Electrical
        an Electrical object
    output : Output
        an Output object
    """

    N0 = output.elec.N0
    omega = 2 * pi * N0 / 60

    P = output.elec.Pem_av_ref
    losses = output.elec.Pj_losses  # TODO update since there may also be other losses

    Tem_av_ref = (P - losses) / omega

    output.elec.Tem_av_ref = Tem_av_ref



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/Mesh/Interpolation/Interpolation.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/pyleecan/tree/master/pyleecan/Methods/Mesh/Interpolation
"""

from os import linesep
from sys import getsizeof
from logging import getLogger
from ._check import check_var, raise_
from ..Functions.get_logger import get_logger
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from ._frozen import FrozenClass

from ._check import InitUnKnowClassError
from .RefCell import RefCell
from .GaussPoint import GaussPoint
from .ScalarProduct import ScalarProduct


class Interpolation(FrozenClass):
    """Store shape functions"""

    VERSION = 1

    # save and copy methods are available in all object
    save = save
    copy = copy
    # get_logger method is available in all object
    get_logger = get_logger

    def __init__(
        self,
        ref_cell=None,
        gauss_point=None,
        scalar_product=None,
        init_dict=None,
        init_str=None,
    ):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for pyleecan type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for pyleecan Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "ref_cell" in list(init_dict.keys()):
                ref_cell = init_dict["ref_cell"]
            if "gauss_point" in list(init_dict.keys()):
                gauss_point = init_dict["gauss_point"]
            if "scalar_product" in list(init_dict.keys()):
                scalar_product = init_dict["scalar_product"]
        # Set the properties (value check and convertion are done in setter)
        self.parent = None
        self.ref_cell = ref_cell
        self.gauss_point = gauss_point
        self.scalar_product = scalar_product

        # The class is frozen, for now it's impossible to add new properties
        self._freeze()

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        Interpolation_str = ""
        if self.parent is None:
            Interpolation_str += "parent = None " + linesep
        else:
            Interpolation_str += (
                "parent = " + str(type(self.parent)) + " object" + linesep
            )
        if self.ref_cell is not None:
            tmp = self.ref_cell.__str__().replace(linesep, linesep + "\t").rstrip("\t")
            Interpolation_str += "ref_cell = " + tmp
        else:
            Interpolation_str += "ref_cell = None" + linesep + linesep
        if self.gauss_point is not None:
            tmp = (
                self.gauss_point.__str__().replace(linesep, linesep + "\t").rstrip("\t")
            )
            Interpolation_str += "gauss_point = " + tmp
        else:
            Interpolation_str += "gauss_point = None" + linesep + linesep
        if self.scalar_product is not None:
            tmp = (
                self.scalar_product.__str__()
                .replace(linesep, linesep + "\t")
                .rstrip("\t")
            )
            Interpolation_str += "scalar_product = " + tmp
        else:
            Interpolation_str += "scalar_product = None" + linesep + linesep
        return Interpolation_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False
        if other.ref_cell != self.ref_cell:
            return False
        if other.gauss_point != self.gauss_point:
            return False
        if other.scalar_product != self.scalar_product:
            return False
        return True

    def compare(self, other, name="self"):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()
        if (other.ref_cell is None and self.ref_cell is not None) or (
            other.ref_cell is not None and self.ref_cell is None
        ):
            diff_list.append(name + ".ref_cell None mismatch")
        elif self.ref_cell is not None:
            diff_list.extend(
                self.ref_cell.compare(other.ref_cell, name=name + ".ref_cell")
            )
        if (other.gauss_point is None and self.gauss_point is not None) or (
            other.gauss_point is not None and self.gauss_point is None
        ):
            diff_list.append(name + ".gauss_point None mismatch")
        elif self.gauss_point is not None:
            diff_list.extend(
                self.gauss_point.compare(other.gauss_point, name=name + ".gauss_point")
            )
        if (other.scalar_product is None and self.scalar_product is not None) or (
            other.scalar_product is not None and self.scalar_product is None
        ):
            diff_list.append(name + ".scalar_product None mismatch")
        elif self.scalar_product is not None:
            diff_list.extend(
                self.scalar_product.compare(
                    other.scalar_product, name=name + ".scalar_product"
                )
            )
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object
        S += getsizeof(self.ref_cell)
        S += getsizeof(self.gauss_point)
        S += getsizeof(self.scalar_product)
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        Interpolation_dict = dict()
        if self.ref_cell is None:
            Interpolation_dict["ref_cell"] = None
        else:
            Interpolation_dict["ref_cell"] = self.ref_cell.as_dict(**kwargs)
        if self.gauss_point is None:
            Interpolation_dict["gauss_point"] = None
        else:
            Interpolation_dict["gauss_point"] = self.gauss_point.as_dict(**kwargs)
        if self.scalar_product is None:
            Interpolation_dict["scalar_product"] = None
        else:
            Interpolation_dict["scalar_product"] = self.scalar_product.as_dict(**kwargs)
        # The class name is added to the dict for deserialisation purpose
        Interpolation_dict["__class__"] = "Interpolation"
        return Interpolation_dict

    def _set_None(self):
        """Set all the properties to None (except pyleecan object)"""

        if self.ref_cell is not None:
            self.ref_cell._set_None()
        if self.gauss_point is not None:
            self.gauss_point._set_None()
        if self.scalar_product is not None:
            self.scalar_product._set_None()

    def _get_ref_cell(self):
        """getter of ref_cell"""
        return self._ref_cell

    def _set_ref_cell(self, value):
        """setter of ref_cell"""
        if isinstance(value, str):  # Load from file
            value = load_init_dict(value)[1]
        if isinstance(value, dict) and "__class__" in value:
            class_obj = import_class(
                "pyleecan.Classes", value.get("__class__"), "ref_cell"
            )
            value = class_obj(init_dict=value)
        elif type(value) is int and value == -1:  # Default constructor
            value = RefCell()
        check_var("ref_cell", value, "RefCell")
        self._ref_cell = value

        if self._ref_cell is not None:
            self._ref_cell.parent = self

    ref_cell = property(
        fget=_get_ref_cell,
        fset=_set_ref_cell,
        doc=u"""

        :Type: RefCell
        """,
    )

    def _get_gauss_point(self):
        """getter of gauss_point"""
        return self._gauss_point

    def _set_gauss_point(self, value):
        """setter of gauss_point"""
        if isinstance(value, str):  # Load from file
            value = load_init_dict(value)[1]
        if isinstance(value, dict) and "__class__" in value:
            class_obj = import_class(
                "pyleecan.Classes", value.get("__class__"), "gauss_point"
            )
            value = class_obj(init_dict=value)
        elif type(value) is int and value == -1:  # Default constructor
            value = GaussPoint()
        check_var("gauss_point", value, "GaussPoint")
        self._gauss_point = value

        if self._gauss_point is not None:
            self._gauss_point.parent = self

    gauss_point = property(
        fget=_get_gauss_point,
        fset=_set_gauss_point,
        doc=u"""

        :Type: GaussPoint
        """,
    )

    def _get_scalar_product(self):
        """getter of scalar_product"""
        return self._scalar_product

    def _set_scalar_product(self, value):
        """setter of scalar_product"""
        if isinstance(value, str):  # Load from file
            value = load_init_dict(value)[1]
        if isinstance(value, dict) and "__class__" in value:
            class_obj = import_class(
                "pyleecan.Classes", value.get("__class__"), "scalar_product"
            )
            value = class_obj(init_dict=value)
        elif type(value) is int and value == -1:  # Default constructor
            value = ScalarProduct()
        check_var("scalar_product", value, "ScalarProduct")
        self._scalar_product = value

        if self._scalar_product is not None:
            self._scalar_product.parent = self

    scalar_product = property(
        fget=_get_scalar_product,
        fset=_set_scalar_product,
        doc=u"""

        :Type: ScalarProduct
        """,
    )



# -*- coding: utf-8 -*-


def comp_height_yoke(self):
    """Compute the yoke height

    Parameters
    ----------
    self : LamSlotMulti
        A LamSlotMulti object

    Returns
    -------
    Hy: float
        yoke height [m]

    """
    H = 0
    for slot in self.slot_list:
        H = max(H, slot.comp_height())
    return self.Rext - self.Rint - H



# -*- coding: utf-8 -*-
from numpy import pi, exp
from ....Classes.Segment import Segment
from ....Classes.SurfLine import SurfLine
from ....Methods.Slot.SlotW61 import S61_WindError


def build_geometry_active(self, Nrad, Ntan, is_simplified=False, alpha=0, delta=0):
    """Split the slot winding area in several zone

    Parameters
    ----------
    self : SlotW61
        A SlotW61 object
    Nrad : int
        Number of radial layer
    Ntan : int
        Number of tangentiel layer
    is_simplified : bool
        boolean to specify if the coincident lines are considered as one
         or different lines (Default value = False)
    alpha : float
        Angle for rotation (Default value = 0) [rad]
    delta : complex
        complex for translation (Default value = 0)

    Returns
    -------
    surf_list: list
        List of surface delimiting the winding zone

    """

    if Nrad != 1 or Ntan != 2:
        raise S61_WindError(
            "Slot 61 can use only for winding with Nrad=1 " + "and Ntan 2"
        )
    self.check()

    # get the name of the lamination
    st = self.get_name_lam()

    [Z1, Z2, Z3, Z4, Z5, Z6, Z7, Z8, Z9, Z10] = self._comp_point_coordinate()

    # Compute the point in the tooth ref
    hsp = pi / self.Zs
    Z4t = Z4 * exp(1j * hsp)
    Z5t = Z5 * exp(1j * hsp)
    Zw1t = Z4t - self.H3
    Zw2t = Z5t + self.H4
    Zw3t = Zw2t + 1j * ((self.W1 - self.W2) / 2 - self.W3)
    Zw4t = Zw1t + 1j * ((self.W1 - self.W2) / 2 - self.W3)

    # Go back to slot ref
    Zw1 = Zw1t * exp(1j * -hsp)
    Zw2 = Zw2t * exp(1j * -hsp)
    Zw3 = Zw3t * exp(1j * -hsp)
    Zw4 = Zw4t * exp(1j * -hsp)
    Zw1s = Zw1.conjugate()
    Zw2s = Zw2.conjugate()
    Zw3s = Zw3.conjugate()
    Zw4s = Zw4.conjugate()
    Ref1 = (Zw1 + Zw2 + Zw3 + Zw4) / 4
    Ref2 = (Zw1s + Zw2s + Zw3s + Zw4s) / 4

    # Create the surfaces
    surf_list = list()
    wind1 = [Segment(Zw3, Zw4)]
    wind2 = [Segment(Zw3s, Zw4s)]
    if (is_simplified and self.W3 > 0) or not is_simplified:
        wind1.append(Segment(Zw4, Zw1))
        wind2.append(Segment(Zw4s, Zw1s))
    if not is_simplified:
        wind1.append(Segment(Zw1, Zw2))
        wind2.append(Segment(Zw1s, Zw2s))
    if (is_simplified and self.W4 > 0) or not is_simplified:
        wind1.append(Segment(Zw2, Zw3))
        wind2.append(Segment(Zw2s, Zw3s))

    surf_list.append(
        SurfLine(line_list=wind1, label="Wind_" + st + "_R0_T0_S0", point_ref=Ref1)
    )
    surf_list.append(
        SurfLine(line_list=wind2, label="Wind_" + st + "_R0_T1_S0", point_ref=Ref2)
    )

    # Rotate and translate the surfaces
    for surf in surf_list:
        surf.rotate(alpha)
        surf.translate(delta)

    return surf_list



# -*- coding: utf-8 -*-
from numpy import pi, exp

from ....Classes.Circle import Circle
from ....Classes.SurfLine import SurfLine
from ....Classes.SurfRing import SurfRing
from ....Classes.Arc1 import Arc1
from ....Classes.Segment import Segment


def build_geometry(self, sym=1, alpha=0, delta=0, is_simplified=False):
    """Build the geometry of the LamHole object

    Parameters
    ----------
    self : LamHole
        The LamHole to build in surface
    sym : int
        Symmetry factor (1= full machine, 2= half of the machine...)
    alpha : float
        Angle for rotation [rad]
    delta : complex
        Complex value for translation
    is_simplified: bool
        True to avoid line superposition

    Returns
    -------
    surf_list : list
        list of surfaces needed to draw the lamination

    """

    # Lamination label
    if self.is_stator:
        label = "Lamination_Stator"
    else:
        label = "Lamination_Rotor"

    if self.is_internal:
        Ryoke = self.Rint
        Rbore = self.Rext
        label1 = "Ext"
        label2 = "Int"
    else:
        Ryoke = self.Rext
        Rbore = self.Rint
        label2 = "Ext"
        label1 = "Int"

    label_bore = label + "_Bore_Radius_" + label1
    label_yoke = label + "_Yoke_Radius_" + label2

    ref_point = self.comp_radius_mid_yoke() * exp(1j * pi / sym)

    surf_list = list()
    # Lamination surface(s)
    if sym == 1:  # Complete lamination
        if self.bore is None:
            bore_surf = SurfLine(
                line_list=self.get_bore_line(0, 2 * pi, label=label_bore),
                label=label_bore,
                point_ref=ref_point,
            )
        else:
            bore_surf = SurfLine(
                line_list=self.bore.get_bore_line(label=label_bore),
                label=label_bore,
                point_ref=ref_point,
            )
        yoke_surf = Circle(
            radius=Ryoke, label=label_yoke, point_ref=0, center=0, line_label=label_yoke
        )
        if Ryoke > 0:
            if self.is_internal:
                surf_list.append(
                    SurfRing(
                        out_surf=bore_surf,
                        in_surf=yoke_surf,
                        label=label,
                        point_ref=ref_point,
                    )
                )
            else:
                surf_list.append(
                    SurfRing(
                        out_surf=yoke_surf,
                        in_surf=bore_surf,
                        label=label,
                        point_ref=ref_point,
                    )
                )
        else:
            surf_list.append(bore_surf)

    else:  # Symmetry lamination
        alpha_begin = 0
        alpha_end = 2 * pi / sym
        begin = Rbore * exp(1j * alpha_begin)
        end = Rbore * exp(1j * alpha_end)
        Z_begin = Ryoke * exp(1j * alpha_begin)
        Z_end = Ryoke * exp(1j * alpha_end)
        line_list = [Segment(Z_begin, begin, label=label + "_Yoke_Side_Right")]
        bore_line = self.get_bore_line(alpha_begin, alpha_end, label=label_bore)
        for line in bore_line:
            line_list.append(line)
        line_list.append(Segment(end, Z_end, label=label + "_Yoke_Side_Left"))
        if Ryoke > 0:
            line_list.append(
                Arc1(
                    begin=Z_end,
                    end=Z_begin,
                    radius=-Ryoke,
                    is_trigo_direction=False,
                    label=label_yoke,
                )
            )
        surf_list.append(
            SurfLine(line_list=line_list, label=label_bore, point_ref=ref_point)
        )

    # Holes surface(s)
    for hole in self.hole:
        Zh = hole.Zh
        assert (Zh % sym) == 0, (
            "ERROR, Wrong symmetry for "
            + label
            + " "
            + str(Zh)
            + " holes and sym="
            + str(sym)
        )  # For now only
        angle = 2 * pi / Zh
        # Create the first hole surface(s)
        surf_hole = hole.build_geometry(alpha=pi / Zh)

        # Copy the hole for Zh / sym
        for ii in range(Zh // sym):
            for surf in surf_hole:
                new_surf = type(surf)(init_dict=surf.as_dict())
                if "Magnet" in surf.label and ii % 2 != 0:  # if the surf is Magnet
                    # Changing the pole of the magnet (before reference number )
                    new_surf.label = new_surf.label[:-10] + "S" + new_surf.label[-9:]
                if "Hole" in surf.label:
                    # changing the hole or magnet reference number
                    new_surf.label = new_surf.label[:-1] + str(ii)
                new_surf.rotate(ii * angle)
                surf_list.append(new_surf)

    # Apply the transformations
    for surf in surf_list:
        surf.rotate(alpha)
        surf.translate(delta)

    # Adding the ventilation surfaces
    for vent in self.axial_vent:
        surf_list += vent.build_geometry(
            sym=sym, alpha=alpha, delta=delta, is_stator=self.is_stator
        )

    return surf_list



import win32com.client
import os
from six import string_types
import pythoncom
from math import exp


class _FEMMHandler(object):
    """
    Object to handle FEMM
    enable to start FEMM and run commands in it
    """

    def __init__(self, HandleToFEMM=None, init_dict=None):
        object.__init__(self)
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "HandleToFEMM" in list(init_dict.keys()):
                HandleToFEMM = init_dict["HandleToFEMM"]
        self.HandleToFEMM = HandleToFEMM
        pythoncom.CoInitialize()

    def as_dict(self, **kwargs):
        _FEMMHandler_dict = dict()
        _FEMMHandler_dict["__class__"] = "_FEMMHandler"
        _FEMMHandler_dict["HandleToFEMM"] = None
        return _FEMMHandler_dict

    def fixpath(self, myPath):
        return myPath.replace("\\", "/").replace("//", "/")

    def openfemm(self, *arg):
        if self.HandleToFEMM != None:
            raise Exception("An instance FEMM is already open")

        self.HandleToFEMM = win32com.client.Dispatch("femm.ActiveFEMM")

        # Call femm
        self.callfemm(
            "setcurrentdirectory(" + self.quote(self.fixpath(os.getcwd())) + ")"
        )
        if len(arg) == 0:
            self.main_restore()
            return
        if arg[0] == 0:
            self.main_restore()
            return

    def closefemm(self):
        self.HandleToFEMM = None  # no reference to the handler, femm is closed

    def callfemm(self, myString):
        HandleToFEMM = self.HandleToFEMM
        x = (
            HandleToFEMM.mlab2femm(myString)
            .replace("[ ", "[")
            .replace(" ]", "]")
            .replace(" ", ",")
            .replace("I", "1j")
        )
        if len(x) == 0:
            x = []
        elif x[0] == "e":
            ErrorMsg = x.replace(",", " ").replace("1j", "I")
            raise Exception(ErrorMsg)
        else:
            x = eval(x)
        if len(x) == 1:
            x = x[0]
        return x

    def main_restore(self):
        self.callfemm("main_restore()")

    def num(self, n):
        return str(n).replace("j", "*I")

    def numc(self, n):
        return self.num(n) + ","

    def quote(self, myString):
        return '"' + myString + '"'

    def quotec(self, myString):
        return '"' + myString + '",'

    def doargs(self, *arg):
        if len(arg) == 0:
            return "()"
        x = "("
        for k in range(0, len(arg)):
            y = arg[k]
            if isinstance(y, string_types):
                x = x + '"' + y + '"'
            else:
                x = x + str(y)
            if k == (len(arg) - 1):
                x = x + ")"
            else:
                x = x + ","
        return x

    def newdocument(self, k):
        self.callfemm("newdocument(" + self.num(k) + ")")

    def AWG(self, awg):
        return 8.2514694 * exp(-0.115943 * awg)

    def callfemm_noeval(self, myString):
        HandleToFEMM = self.HandleToFEMM
        HandleToFEMM.mlab2femm(myString)

    def ci_addarc(self, *arg):
        self.callfemm("ci_addarc" + self.doargs(*arg))

    def ci_addblocklabel(self, *arg):
        self.callfemm("ci_addblocklabel" + self.doargs(*arg))

    def ci_addboundprop(self, *arg):
        self.callfemm("ci_addboundprop" + self.doargs(*arg))

    def ci_addconductorprop(self, *arg):
        self.callfemm("ci_addconductorprop" + self.doargs(*arg))

    def ci_addmaterial(self, *arg):
        self.callfemm("ci_addmaterial" + self.doargs(*arg))

    def ci_addnode(self, *arg):
        self.callfemm("ci_addnode" + self.doargs(*arg))

    def ci_addpointprop(self, *arg):
        self.callfemm("ci_addpointprop" + self.doargs(*arg))

    def ci_addsegment(self, *arg):
        self.callfemm("ci_addsegment" + self.doargs(*arg))

    def ci_analyze(self, *arg):
        self.callfemm("ci_analyze" + self.doargs(*arg))

    def ci_analyse(self, n):
        self.ci_analyze(n)

    def ci_attachdefault(self):
        self.callfemm("ci_attachdefault()")

    def ci_clearselected(self):
        self.callfemm("ci_clearselected()")

    def ci_cleartkpoints(self, n):
        self.callfemm("ci_cleartkpoints(" + self.quote(n) + ")")

    def ci_close(self):
        self.callfemm("ci_close()")

    def ci_copyrotate(self, *arg):
        self.callfemm("ci_copyrotate" + self.doargs(*arg))

    def ci_copyrotate2(self, *arg):
        self.callfemm("ci_copyrotate" + self.doargs(*arg))

    def ci_copytranslate(self, *arg):
        self.callfemm("ci_copytranslate" + self.doargs(*arg))

    def ci_createmesh(self):
        return self.callfemm("ci_createmesh()")

    def ci_createradius(self, x, y, r):
        self.callfemm(
            "ci_createradius(" + self.numc(x) + self.numc(y) + self.num(r) + ")"
        )

    def ci_deleteboundprop(self, n):
        self.callfemm("ci_deleteboundprop(" + self.quote(n) + ")")

    def ci_deleteconductor(self, n):
        self.callfemm("ci_deleteconductor(" + self.quote(n) + ")")

    def ci_deletematerial(self, n):
        self.callfemm("ci_deletematerial(" + self.quote(n) + ")")

    def ci_deletepointprop(self, n):
        self.callfemm("ci_deletepointprop(" + self.quote(n) + ")")

    def ci_deleteselected(self):
        self.callfemm("ci_deleteselected()")

    def ci_deleteselectedarcsegments(self):
        self.callfemm("ci_deleteselectedarcsegments()")

    def ci_deleteselectedlabels(self):
        self.callfemm("ci_deleteselectedlabels()")

    def ci_deleteselectednodes(self):
        self.callfemm("ci_deleteselectednodes()")

    def ci_deleteselectedsegments(self):
        self.callfemm("ci_deleteselectedsegments()")

    def ci_detachdefault(self):
        self.callfemm("ci_detachdefault()")

    def ci_drawarc(self, x1, y1, x2, y2, angle, maxseg):
        self.ci_addnode(x1, y1)
        self.ci_addnode(x2, y2)
        self.ci_addarc(x1, y1, x2, y2, angle, maxseg)

    def ci_drawline(self, x1, y1, x2, y2):
        self.ci_addnode(x1, y1)
        self.ci_addnode(x2, y2)
        self.ci_addsegment(x1, y1, x2, y2)

    def ci_drawrectangle(self, x1, y1, x2, y2):
        self.ci_drawline(x1, y1, x2, y1)
        self.ci_drawline(x2, y1, x2, y2)
        self.ci_drawline(x2, y2, x1, y2)
        self.ci_drawline(x1, y2, x1, y1)

    def ci_getmaterial(self, matname):
        self.callfemm("ci_getmaterial(" + self.quote(matname) + ")")

    def ci_hidegrid(self):
        self.callfemm("ci_hidegrid()")

    def ci_hidenames(self):
        self.callfemm("ci_shownames(0)")

    def ci_loadsolution(self):
        self.callfemm("ci_loadsolution()")

    def ci_makeABC(self, *args):
        self.callfemm("ci_makeABC" + self.doargs(*args))

    def ci_maximize(self):
        self.callfemm("ci_maximize()")

    def ci_minimize(self):
        self.callfemm("ci_minimize()")

    def ci_mirror(self, x1, x2, y1, y2):
        self.callfemm(
            "ci_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def ci_mirror2(self, x1, x2, y1, y2, editaction):
        self.callfemm(
            "ci_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.num(editaction)
            + ")"
        )

    def ci_modifyboundprop(self, *arg):
        self.callfemm("ci_modifyboundprop" + self.doargs(*arg))

    def ci_modifyconductorprop(self, *arg):
        self.callfemm("ci_modifyconductorprop" + self.doargs(*arg))

    def ci_modifymaterial(self, *arg):
        self.callfemm("ci_modifymaterial" + self.doargs(*arg))

    def ci_modifypointprop(self, *arg):
        self.callfemm("ci_modifypointprop" + self.doargs(*arg))

    def ci_moverotate(self, bx, by, shiftangle):
        self.callfemm(
            "ci_moverotate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(shiftangle)
            + ")"
        )

    def ci_movetranslate(self, bx, by):
        self.callfemm("ci_movetranslate(" + self.numc(bx) + self.num(by) + ")")

    def ci_movetranslate2(self, bx, by, editaction):
        self.callfemm(
            "ci_movetranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(editaction)
            + ")"
        )

    def ci_probdef(self, *arg):
        self.callfemm("ci_probdef" + self.doargs(*arg))

    def ci_purgemesh(self):
        self.callfemm("ci_purgemesh()")

    def ci_readdxf(self, docname):
        self.callfemm("print(ci_readdxf(" + self.quote(docname) + "))")

    def ci_refreshview(self):
        self.callfemm("ci_refreshview()")

    def ci_resize(self, nWidth, nHeight):
        self.callfemm("ci_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def ci_restore(self):
        self.callfemm("ci_restore()")

    def ci_saveas(self, fn):
        self.callfemm("ci_saveas(" + self.quote(self.fixpath(fn)) + ")")

    def ci_savebitmap(self, n):
        self.callfemm("ci_savebitmap(" + self.quote(n) + ")")

    def ci_savedxf(self, docname):
        self.callfemm_noeval("ci_savedxf(" + self.quote(docname) + ")")

    def ci_savemetafile(self, n):
        self.callfemm("ci_savemetafile(" + self.quote(n) + ")")

    def ci_scale(self, bx, by, sc):
        self.callfemm("ci_scale(" + self.numc(bx) + self.numc(by) + self.numc(sc) + ")")

    def ci_scale2(self, bx, by, sc, ea):
        self.callfemm(
            "ci_scale("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(sc)
            + self.num(ea)
            + ")"
        )

    def ci_selectarcsegment(self, x, y):
        return self.callfemm("ci_selectarcsegment(" + self.numc(x) + self.num(y) + ")")

    def ci_selectcircle(self, *arg):
        return self.callfemm("ci_selectcircle" + self.doargs(*arg))

    def ci_selectgroup(self, gr):
        return self.callfemm("ci_selectgroup(" + self.num(gr) + ")")

    def ci_selectlabel(self, x, y):
        return self.callfemm("ci_selectlabel(" + self.numc(x) + self.num(y) + ")")

    def ci_selectnode(self, x, y):
        return self.callfemm("ci_selectnode(" + self.numc(x) + self.num(y) + ")")

    def ci_selectrectangle(self, *arg):
        self.callfemm("ci_selectrectangle" + self.doargs(*arg))

    def ci_selectsegment(self, x, y):
        return self.callfemm("ci_selectsegment(" + self.numc(x) + self.num(y) + ")")

    def ci_setarcsegmentprop(self, maxsegdeg, propname, hide, group, incond):
        self.callfemm(
            "ci_setarcsegmentprop("
            + self.numc(maxsegdeg)
            + self.quotec(propname)
            + self.numc(hide)
            + self.numc(group)
            + self.quote(incond)
            + ")"
        )

    def ci_setblockprop(self, blockname, automesh, meshsize, group):
        self.callfemm(
            "ci_setblockprop("
            + self.quotec(blockname)
            + self.numc(automesh)
            + self.numc(meshsize)
            + self.num(group)
            + ")"
        )

    def ci_seteditmode(self, editmode):
        self.callfemm("ci_seteditmode(" + self.quote(editmode) + ")")

    def ci_setfocus(self, docname):
        self.callfemm("ci_setfocus(" + self.quote(docname) + ")")

    def ci_setgrid(self, density, ptype):
        self.callfemm("ci_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def ci_setgroup(self, n):
        return self.callfemm("ci_setgroup(" + self.num(n) + ")")

    def ci_setnodeprop(self, nodeprop, groupno, incond):
        self.callfemm(
            "ci_setnodeprop("
            + self.quotec(nodeprop)
            + self.numc(groupno)
            + self.quote(incond)
            + ")"
        )

    def ci_setsegmentprop(self, pn, es, am, hi, gr, incond):
        self.callfemm(
            "ci_setsegmentprop("
            + self.quotec(pn)
            + self.numc(es)
            + self.numc(am)
            + self.numc(hi)
            + self.numc(gr)
            + self.quote(incond)
            + ")"
        )

    def ci_showgrid(self):
        self.callfemm("ci_showgrid()")

    def ci_showmesh(self):
        self.callfemm("ci_showmesh()")

    def ci_shownames(self):
        self.callfemm("ci_shownames(1)")

    def ci_smartmesh(self, n):
        self.callfemm("ci_smartmesh(" + self.num(n) + ")")

    def ci_snapgridoff(self):
        self.callfemm('ci_gridsnap("off")')

    def ci_snapgridon(self):
        self.callfemm('ci_gridsnap("on")')

    def ci_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "ci_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def ci_zoomin(self):
        self.callfemm("ci_zoomin()")

    def ci_zoomnatural(self):
        self.callfemm("ci_zoomnatural()")

    def ci_zoomout(self):
        self.callfemm("ci_zoomout()")

    def co_addcontour(self, x, y):
        self.callfemm("co_addcontour(" + self.numc(x) + self.num(y) + ")")

    def co_bendcontour(self, tta, dtta):
        self.callfemm("co_bendcontour(" + self.numc(tta) + self.num(dtta) + ")")

    def co_blockintegral(self, ptype):
        return self.callfemm("co_blockintegral(" + self.num(ptype) + ")")

    def co_clearblock(self):
        self.callfemm("co_clearblock()")

    def co_clearcontour(self):
        self.callfemm("co_clearcontour()")

    def co_close(self):
        self.callfemm("co_close()")

    def co_getconductorproperties(self, pname):
        return self.callfemm("co_getconductorproperties(" + self.quote(pname) + ")")

    def co_gete(self, x, y):
        z = self.co_getpointvalues(x, y)
        return [z[5], z[6]]

    def co_getelement(self, n):
        return self.callfemm("co_getelement(" + self.num(n) + ")")

    def co_getj(self, x, y):
        z = self.co_getpointvalues(x, y)
        return [z[1], z[2]]

    def co_getk(self, x, y):
        z = self.co_getpointvalues(x, y)
        return [z[3], z[4]]

    def co_getnode(self, n):
        return self.callfemm("co_getnode(" + self.num(n) + ")")

    def co_getpointvalues(self, x, y):
        return self.callfemm("co_getpointvalues(" + self.numc(x) + self.num(y) + ")")

    def co_getprobleminfo(self):
        return self.callfemm("co_getprobleminfo()")

    def co_getv(self, x, y):
        z = self.co_getpointvalues(x, y)
        return z[0]

    def co_groupselectblock(self, *arg):
        self.callfemm("co_groupselectblock" + self.doargs(*arg))

    def co_hidecontourplot(self):
        self.callfemm("co_hidecontourplot()")

    def co_hidedensityplot(self):
        self.callfemm("co_hidedensityplot()")

    def co_hidegrid(self):
        self.callfemm("co_hidegrid()")

    def co_hidemesh(self):
        self.callfemm("co_hidemesh()")

    def co_hidenames(self):
        self.callfemm("co_shownames(0)")

    def co_hidepoints(self):
        self.callfemm("co_hidepoints()")

    def co_lineintegral(self, ptype):
        return self.callfemm("co_lineintegral(" + self.num(ptype) + ")")

    def co_makeplot(self, *arg):
        self.callfemm("co_makeplot" + self.doargs(*arg))

    def co_maximize(self):
        self.callfemm("co_maximize()")

    def co_minimize(self):
        self.callfemm("co_minimize()")

    def co_numelements(self):
        return self.callfemm("co_numelements()")

    def co_numnodes(self):
        return self.callfemm("co_numnodes()")

    def co_refreshview(self):
        self.callfemm("co_refreshview()")

    def co_reload(self):
        self.callfemm("co_reload()")

    def co_resize(self, nWidth, nHeight):
        self.callfemm("co_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def co_restore(self):
        self.callfemm("co_restore()")

    def co_savebitmap(self, fn):
        self.callfemm("co_savebitmap(" + self.quote(self.fixpath(fn)) + ")")

    def co_savemetafile(self, fn):
        self.callfemm("co_savemetafile(" + self.quote(self.fixpath(fn)) + ")")

    def co_selectblock(self, x, y):
        self.callfemm("co_selectblock(" + self.numc(x) + self.num(y) + ")")

    def co_selectpoint(self, x, y):
        self.callfemm("co_selectpoint(" + self.numc(x) + self.num(y) + ")")

    def co_seteditmode(self, mode):
        self.callfemm("co_seteditmode(" + self.quote(mode) + ")")

    def co_setgrid(self, density, ptype):
        self.callfemm("co_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def co_showcontourplot(self, numcontours, al, au):
        self.callfemm(
            "co_showcontourplot("
            + self.numc(numcontours)
            + self.numc(al)
            + self.num(au)
            + ")"
        )

    def co_showdensityplot(self, legend, gscale, ptype, bu, bl):
        self.callfemm(
            "co_showdensityplot("
            + self.numc(legend)
            + self.numc(gscale)
            + self.numc(ptype)
            + self.numc(bu)
            + self.num(bl)
            + ")"
        )

    def co_showgrid(self):
        self.callfemm("co_showgrid()")

    def co_showmesh(self):
        self.callfemm("co_showmesh()")

    def co_shownames(self):
        self.callfemm("co_shownames(1)")

    def co_showpoints(self):
        self.callfemm("co_showpoints()")

    def co_showvectorplot(self, *arg):
        self.callfemm("co_showvectorplot" + self.doargs(*arg))

    def co_smooth(self, flag):
        self.callfemm("co_smooth(" + self.quote(flag) + ")")

    def co_smoothoff(self):
        self.callfemm('co_smooth("off")')

    def co_smoothon(self):
        self.callfemm('co_smooth("on")')

    def co_snapgrid(self, flag):
        self.callfemm("co_gridsnap(" + self.quote(flag) + ")")

    def co_snapgridoff(self):
        self.callfemm('co_gridsnap("off")')

    def co_snapgridon(self):
        self.callfemm('co_gridsnap("on")')

    def co_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "co_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def co_zoomin(self):
        self.callfemm("co_zoomin()")

    def co_zoomnatural(self):
        self.callfemm("co_zoomnatural()")

    def co_zoomout(self):
        self.callfemm("co_zoomout()")

    def complex2str(self, x):
        return str(x)

    def create(self, n):
        self.callfemm("create(" + self.num(n) + ")")

    def ei_addarc(self, x1, y1, x2, y2, angle, maxseg):
        self.callfemm(
            "ei_addarc("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.numc(angle)
            + self.num(maxseg)
            + ")"
        )

    def ei_addblocklabel(self, x, y):
        self.callfemm("ei_addblocklabel(" + self.numc(x) + self.num(y) + ")")

    def ei_addboundprop(self, pname, vs, qs, c0, c1, fmt):
        self.callfemm(
            "ei_addboundprop("
            + self.quotec(pname)
            + self.numc(vs)
            + self.numc(qs)
            + self.numc(c0)
            + self.numc(c1)
            + self.num(fmt)
            + ")"
        )

    def ei_addconductorprop(self, pname, vc, qc, ptype):
        self.callfemm(
            "ei_addconductorprop("
            + self.quotec(pname)
            + self.numc(vc)
            + self.numc(qc)
            + self.num(ptype)
            + ")"
        )

    def ei_addmaterial(self, pname, ex, ey, qv):
        self.callfemm(
            "ei_addmaterial("
            + self.quotec(pname)
            + self.numc(ex)
            + self.numc(ey)
            + self.num(qv)
            + ")"
        )

    def ei_addnode(self, x, y):
        self.callfemm("ei_addnode(" + self.numc(x) + self.num(y) + ")")

    def ei_addpointprop(self, pname, vp, qp):
        self.callfemm(
            "ei_addpointprop(" + self.quotec(pname) + self.numc(vp) + self.num(qp) + ")"
        )

    def ei_addsegment(self, x1, y1, x2, y2):
        self.callfemm(
            "ei_addsegment("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def ei_analyze(self, *arg):
        self.callfemm("ei_analyze" + self.doargs(*arg))

    def ei_analyse(self, n):
        self.ei_analyze(n)

    def ei_attachdefault(self):
        self.callfemm("ei_attachdefault()")

    def ei_attachouterspace(self):
        self.callfemm("ei_attachouterspace()")

    def ei_clearselected(self):
        self.callfemm("ei_clearselected()")

    def ei_close(self):
        self.callfemm("ei_close()")

    def ei_copyrotate(self, bx, by, angle, copies):
        self.callfemm(
            "ei_copyrotate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(angle)
            + self.num(copies)
            + ")"
        )

    def ei_copyrotate2(self, bx, by, angle, copies, editaction):
        self.callfemm(
            "ei_copyrotate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(angle)
            + self.numc(copies)
            + self.num(editaction)
            + ")"
        )

    def ei_copytranslate(self, bx, by, copies):
        self.callfemm(
            "ei_copytranslate(" + self.numc(bx) + self.numc(by) + self.num(copies) + ")"
        )

    def ei_copytranslate2(self, bx, by, copies, editaction):
        self.callfemm(
            "ei_copytranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(copies)
            + self.num(editaction)
            + ")"
        )

    def ei_createmesh(self):
        return self.callfemm("ei_createmesh()")

    def ei_createradius(self, x, y, r):
        self.callfemm(
            "ei_createradius(" + self.numc(x) + self.numc(y) + self.num(r) + ")"
        )

    def ei_defineouterspace(self, Zo, Ro, Ri):
        self.callfemm(
            "ei_defineouterspace(" + self.numc(Zo) + self.numc(Ro) + self.num(Ri) + ")"
        )

    def ei_deleteboundprop(self, n):
        self.callfemm("ei_deleteboundprop(" + self.quote(n) + ")")

    def ei_deleteconductor(self, n):
        self.callfemm("ei_deleteconductor(" + self.quote(n) + ")")

    def ei_deletematerial(self, n):
        self.callfemm("ei_deletematerial(" + self.quote(n) + ")")

    def ei_deletepointprop(self, n):
        self.callfemm("ei_deletepointprop(" + self.quote(n) + ")")

    def ei_deleteselected(self):
        self.callfemm("ei_deleteselected()")

    def ei_deleteselectedarcsegments(self):
        self.callfemm("ei_deleteselectedarcsegments()")

    def ei_deleteselectedlabels(self):
        self.callfemm("ei_deleteselectedlabels()")

    def ei_deleteselectednodes(self):
        self.callfemm("ei_deleteselectednodes()")

    def ei_deleteselectedsegments(self):
        self.callfemm("ei_deleteselectedsegments()")

    def ei_detachdefault(self):
        self.callfemm("ei_detachdefault()")

    def ei_detachouterspace(self):
        self.callfemm("ei_detachouterspace()")

    def ei_drawarc(self, x1, y1, x2, y2, angle, maxseg):
        self.ei_addnode(x1, y1)
        self.ei_addnode(x2, y2)
        self.ei_addarc(x1, y1, x2, y2, angle, maxseg)

    def ei_drawline(self, x1, y1, x2, y2):
        self.ei_addnode(x1, y1)
        self.ei_addnode(x2, y2)
        self.ei_addsegment(x1, y1, x2, y2)

    def ei_drawrectangle(self, x1, y1, x2, y2):
        self.ei_drawline(x1, y1, x2, y1)
        self.ei_drawline(x2, y1, x2, y2)
        self.ei_drawline(x2, y2, x1, y2)
        self.ei_drawline(x1, y2, x1, y1)

    def ei_getmaterial(self, matname):
        self.callfemm("ei_getmaterial(" + self.quote(matname) + ")")

    def ei_hidegrid(self):
        self.callfemm("ei_hidegrid()")

    def ei_hidenames(self):
        self.callfemm("ei_shownames(0)")

    def ei_loadsolution(self):
        self.callfemm("ei_loadsolution()")

    def ei_makeABC(self, *arg):
        self.callfemm("ei_makeABC" + self.doargs(*arg))

    def ei_maximize(self):
        self.callfemm("ei_maximize()")

    def ei_minimize(self):
        self.callfemm("ei_minimize()")

    def ei_mirror(self, x1, y1, x2, y2):
        self.callfemm(
            "ei_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def ei_mirror2(self, x1, y1, x2, y2, editaction):
        self.callfemm(
            "ei_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.num(editaction)
            + ")"
        )

    def ei_modifyboundprop(self, *arg):
        self.callfemm("ei_modifyboundprop" + self.doargs(*arg))

    def ei_modifyconductorprop(self, *arg):
        self.callfemm("ei_modifyconductorprop" + self.doargs(*arg))

    def ei_modifymaterial(self, *arg):
        self.callfemm("ei_modifymaterial" + self.doargs(*arg))

    def ei_modifypointprop(self, *arg):
        self.callfemm("ei_modifypointprop" + self.doargs(*arg))

    def ei_moverotate(self, bx, by, shiftangle):
        self.callfemm(
            "ei_moverotate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(shiftangle)
            + ")"
        )

    def ei_movetranslate(self, bx, by):
        self.callfemm("ei_movetranslate(" + self.numc(bx) + self.num(by) + ")")

    def ei_movetranslate2(self, bx, by, editaction):
        self.callfemm(
            "ei_movetranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(editaction)
            + ")"
        )

    def ei_probdef(self, *arg):
        self.callfemm("ei_probdef" + self.doargs(*arg))

    def ei_purgemesh(self):
        self.callfemm("ei_purgemesh()")

    def ei_readdxf(self, docname):
        self.callfemm("print(ei_readdxf(" + self.quote(docname) + "))")

    def ei_refreshview(self):
        self.callfemm("ei_refreshview()")

    def ei_resize(self, nWidth, nHeight):
        self.callfemm("ei_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def ei_restore(self):
        self.callfemm("ei_restore()")

    def ei_saveas(self, fn):
        self.callfemm("ei_saveas(" + self.quote(self.fixpath(fn)) + ")")

    def ei_savebitmap(self, n):
        self.callfemm("ei_savebitmap(" + self.quote(n) + ")")

    def ei_savedxf(self, docname):
        self.callfemm_noeval("ei_savedxf(" + self.quote(docname) + ")")

    def ei_savemetafile(self, n):
        self.callfemm("ei_savemetafile(" + self.quote(n) + ")")

    def ei_scale(self, bx, by, sc):
        self.callfemm("ei_scale(" + self.numc(bx) + self.numc(by) + self.numc(sc) + ")")

    def ei_scale2(self, bx, by, sc, ea):
        self.callfemm(
            "ei_scale("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(sc)
            + self.num(ea)
            + ")"
        )

    def ei_selectarcsegment(self, x, y):
        return self.callfemm("ei_selectarcsegment(" + self.numc(x) + self.num(y) + ")")

    def ei_selectcircle(self, *arg):
        self.callfemm("ei_selectcircle" + self.doargs(*arg))

    def ei_selectgroup(self, gr):
        self.callfemm("ei_selectgroup(" + self.num(gr) + ")")

    def ei_selectlabel(self, x, y):
        return self.callfemm("ei_selectlabel(" + self.numc(x) + self.num(y) + ")")

    def ei_selectnode(self, x, y):
        return self.callfemm("ei_selectnode(" + self.numc(x) + self.num(y) + ")")

    def ei_selectrectangle(self, *arg):
        self.callfemm("ei_selectrectangle" + self.doargs(*arg))

    def ei_selectsegment(self, x, y):
        return self.callfemm("ei_selectsegment(" + self.numc(x) + self.num(y) + ")")

    def ei_setarcsegmentprop(self, maxsegdeg, propname, hide, group, incond):
        self.callfemm(
            "ei_setarcsegmentprop("
            + self.numc(maxsegdeg)
            + self.quotec(propname)
            + self.numc(hide)
            + self.numc(group)
            + self.quote(incond)
            + ")"
        )

    def ei_setblockprop(self, blockname, automesh, meshsize, group):
        self.callfemm(
            "ei_setblockprop("
            + self.quotec(blockname)
            + self.numc(automesh)
            + self.numc(meshsize)
            + self.num(group)
            + ")"
        )

    def ei_seteditmode(self, editmode):
        self.callfemm("ei_seteditmode(" + self.quote(editmode) + ")")

    def ei_setfocus(self, docname):
        self.callfemm("ei_setfocus(" + self.quote(docname) + ")")

    def ei_setgrid(self, density, ptype):
        self.callfemm("ei_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def ei_setgroup(self, n):
        return self.callfemm("ei_setgroup(" + self.num(n) + ")")

    def ei_setnodeprop(self, nodeprop, groupno, inconductor):
        self.callfemm(
            "ei_setnodeprop("
            + self.quotec(nodeprop)
            + self.numc(groupno)
            + self.quote(inconductor)
            + ")"
        )

    def ei_setsegmentprop(self, pn, es, am, hi, gr, inconductor):
        self.callfemm(
            "ei_setsegmentprop("
            + self.quotec(pn)
            + self.numc(es)
            + self.numc(am)
            + self.numc(hi)
            + self.numc(gr)
            + self.quote(inconductor)
            + ")"
        )

    def ei_showgrid(self):
        self.callfemm("ei_showgrid()")

    def ei_showmesh(self):
        self.callfemm("ei_showmesh()")

    def ei_shownames(self):
        self.callfemm("ei_shownames(1)")

    def ei_smartmesh(self, n):
        self.callfemm("ei_smartmesh(" + self.num(n) + ")")

    def ei_snapgridoff(self):
        self.callfemm('ei_gridsnap("off")')

    def ei_snapgridon(self):
        self.callfemm('ei_gridsnap("on")')

    def ei_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "ei_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def ei_zoomin(self):
        self.callfemm("ei_zoomin()")

    def ei_zoomnatural(self):
        self.callfemm("ei_zoomnatural()")

    def ei_zoomout(self):
        self.callfemm("ei_zoomout()")

    def eo_addcontour(self, x, y):
        self.callfemm("eo_addcontour(" + self.numc(x) + self.num(y) + ")")

    def eo_bendcontour(self, tta, dtta):
        self.callfemm("eo_bendcontour(" + self.numc(tta) + self.num(dtta) + ")")

    def eo_blockintegral(self, ptype):
        return self.callfemm("eo_blockintegral(" + self.num(ptype) + ")")

    def eo_clearblock(self):
        self.callfemm("eo_clearblock()")

    def eo_clearcontour(self):
        self.callfemm("eo_clearcontour()")

    def eo_close(self):
        self.callfemm("eo_close()")

    def eo_getconductorproperties(self, pname):
        return self.callfemm("eo_getconductorproperties(" + self.quote(pname) + ")")

    def eo_getd(self, x, y):
        z = self.eo_getpointvalues(x, y)
        return [z[1], z[2]]

    def eo_gete(self, x, y):
        z = self.eo_getpointvalues(x, y)
        return [z[3], z[4]]

    def eo_getelement(self, n):
        return self.callfemm("eo_getelement(" + self.num(n) + ")")

    def eo_getenergydensity(self, x, y):
        z = self.eo_getpointvalues(x, y)
        return z[7]

    def eo_getnode(self, n):
        return self.callfemm("eo_getnode(" + self.num(n) + ")")

    def eo_getperm(self, x, y):
        z = self.eo_getpointvalues(x, y)
        return [z[5], z[6]]

    def eo_getpointvalues(self, x, y):
        z = self.callfemm("eo_getpointvalues(" + self.numc(x) + self.num(y) + ")")
        if len(z) > 0:
            return z
        return [0, 0, 0, 0, 0, 0, 0, 0]

    def eo_getprobleminfo(self):
        return self.callfemm("eo_getprobleminfo()")

    def eo_getv(self, x, y):
        z = self.eo_getpointvalues(x, y)
        return z[0]

    def eo_groupselectblock(self, *arg):
        self.callfemm("eo_groupselectblock" + self.doargs(*arg))

    def eo_hidecontourplot(self):
        self.callfemm("eo_hidecontourplot()")

    def eo_hidedensityplot(self):
        self.callfemm("eo_hidedensityplot()")

    def eo_hidegrid(self):
        self.callfemm("eo_hidegrid()")

    def eo_hidemesh(self):
        self.callfemm("eo_hidemesh()")

    def eo_hidenames(self):
        self.callfemm("eo_shownames(0)")

    def eo_hidepoints(self):
        self.callfemm("eo_hidepoints()")

    def eo_lineintegral(self, ptype):
        return self.callfemm("eo_lineintegral(" + self.num(ptype) + ")")

    def eo_makeplot(self, *arg):
        self.callfemm("eo_makeplot" + self.doargs(*arg))

    def eo_maximize(self):
        self.callfemm("eo_maximize()")

    def eo_minimize(self):
        self.callfemm("eo_minimize()")

    def eo_numelements(self):
        return self.callfemm("eo_numelements()")

    def eo_numnodes(self):
        return self.callfemm("eo_numnodes()")

    def eo_refreshview(self):
        self.callfemm("eo_refreshview()")

    def eo_reload(self):
        self.callfemm("eo_reload()")

    def eo_resize(self, nWidth, nHeight):
        self.callfemm("eo_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def eo_restore(self):
        self.callfemm("eo_restore()")

    def eo_savebitmap(self, fn):
        self.callfemm("eo_savebitmap(" + self.quote(self.fixpath(fn)) + ")")

    def eo_savemetafile(self, fn):
        self.callfemm("eo_savemetafile(" + self.quote(self.fixpath(fn)) + ")")

    def eo_selectblock(self, x, y):
        self.callfemm("eo_selectblock(" + self.numc(x) + self.num(y) + ")")

    def eo_selectpoint(self, x, y):
        self.callfemm("eo_selectpoint(" + self.numc(x) + self.num(y) + ")")

    def eo_seteditmode(self, mode):
        self.callfemm("eo_seteditmode(" + self.quote(mode) + ")")

    def eo_setgrid(self, density, ptype):
        self.callfemm("eo_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def eo_showcontourplot(self, numcontours, al, au):
        self.callfemm(
            "eo_showcontourplot("
            + self.numc(numcontours)
            + self.numc(al)
            + self.num(au)
            + ")"
        )

    def eo_showdensityplot(self, legend, gscale, ptype, bu, bl):
        self.callfemm(
            "eo_showdensityplot("
            + self.numc(legend)
            + self.numc(gscale)
            + self.numc(ptype)
            + self.numc(bu)
            + self.num(bl)
            + ")"
        )

    def eo_showgrid(self):
        self.callfemm("eo_showgrid()")

    def eo_showmesh(self):
        self.callfemm("eo_showmesh()")

    def eo_shownames(self):
        self.callfemm("eo_shownames(1)")

    def eo_showpoints(self):
        self.callfemm("eo_showpoints()")

    def eo_showvectorplot(self, *arg):
        self.callfemm("eo_showvectorplot" + self.doargs(*arg))

    def eo_smooth(self, flag):
        self.callfemm("eo_smooth(" + self.quote(flag) + ")")

    def eo_smoothoff(self):
        self.callfemm('eo_smooth("off")')

    def eo_smoothon(self):
        self.callfemm('eo_smooth("on")')

    def eo_snapgrid(self, flag):
        self.callfemm("eo_gridsnap(" + self.quote(flag) + ")")

    def eo_snapgridoff(self):
        self.callfemm('eo_gridsnap("off")')

    def eo_snapgridon(self):
        self.callfemm('eo_gridsnap("on")')

    def eo_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "eo_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def eo_zoomin(self):
        self.callfemm("eo_zoomin()")

    def eo_zoomnatural(self):
        self.callfemm("eo_zoomnatural()")

    def eo_zoomout(self):
        self.callfemm("eo_zoomout()")

    def hi_addarc(self, x1, y1, x2, y2, angle, maxseg):
        self.callfemm(
            "hi_addarc("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.numc(angle)
            + self.num(maxseg)
            + ")"
        )

    def hi_addblocklabel(self, x, y):
        self.callfemm("hi_addblocklabel(" + self.numc(x) + self.num(y) + ")")

    def hi_addboundprop(self, *arg):
        self.callfemm("hi_addboundprop" + self.doargs(*arg))

    def hi_addconductorprop(self, *arg):
        self.callfemm("hi_addconductorprop" + self.doargs(*arg))

    def hi_addmaterial(self, *arg):
        self.callfemm("hi_addmaterial" + self.doargs(*arg))

    def hi_addnode(self, x, y):
        self.callfemm("hi_addnode(" + self.numc(x) + self.num(y) + ")")

    def hi_addpointprop(self, *arg):
        self.callfemm("hi_addpointprop" + self.doargs(*arg))

    def hi_addsegment(self, x1, y1, x2, y2):
        self.callfemm(
            "hi_addsegment("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def hi_addtkpoint(self, name, b, h):
        self.callfemm(
            "hi_addtkpoint(" + self.quotec(name) + self.numc(b) + self.num(h) + ")"
        )

    def hi_analyze(self, *arg):
        self.callfemm("hi_analyze" + self.doargs(*arg))

    def hi_analyse(self, n):
        self.hi_analyze(n)

    def hi_attachdefault(self):
        self.callfemm("hi_attachdefault()")

    def hi_attachouterspace(self):
        self.callfemm("hi_attachouterspace()")

    def hi_clearselected(self):
        self.callfemm("hi_clearselected()")

    def hi_cleartkpoints(self, n):
        self.callfemm("hi_cleartkpoints(" + self.quote(n) + ")")

    def hi_close(self):
        self.callfemm("hi_close()")

    def hi_copyrotate(self, bx, by, angle, copies):
        self.callfemm(
            "hi_copyrotate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(angle)
            + self.num(copies)
            + ")"
        )

    def hi_copyrotate2(self, bx, by, angle, copies, editaction):
        self.callfemm(
            "hi_copyrotate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(angle)
            + self.numc(copies)
            + self.num(editaction)
            + ")"
        )

    def hi_copytranslate(self, bx, by, copies):
        self.callfemm(
            "hi_copytranslate(" + self.numc(bx) + self.numc(by) + self.num(copies) + ")"
        )

    def hi_copytranslate2(self, bx, by, copies, editaction):
        self.callfemm(
            "hi_copytranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(copies)
            + self.num(editaction)
            + ")"
        )

    def hi_createmesh(self):
        return self.callfemm("hi_createmesh()")

    def hi_createradius(self, x, y, r):
        self.callfemm(
            "hi_createradius(" + self.numc(x) + self.numc(y) + self.num(r) + ")"
        )

    def hi_defineouterspace(self, Zo, Ro, Ri):
        self.callfemm(
            "hi_defineouterspace(" + self.numc(Zo) + self.numc(Ro) + self.num(Ri) + ")"
        )

    def hi_deleteboundprop(self, n):
        self.callfemm("hi_deleteboundprop(" + self.quote(n) + ")")

    def hi_deleteconductor(self, n):
        self.callfemm("hi_deleteconductor(" + self.quote(n) + ")")

    def hi_deletematerial(self, n):
        self.callfemm("hi_deletematerial(" + self.quote(n) + ")")

    def hi_deletepointprop(self, n):
        self.callfemm("hi_deletepointprop(" + self.quote(n) + ")")

    def hi_deleteselected(self):
        self.callfemm("hi_deleteselected()")

    def hi_deleteselectedarcsegments(self):
        self.callfemm("hi_deleteselectedarcsegments()")

    def hi_deleteselectedlabels(self):
        self.callfemm("hi_deleteselectedlabels()")

    def hi_deleteselectednodes(self):
        self.callfemm("hi_deleteselectednodes()")

    def hi_deleteselectedsegments(self):
        self.callfemm("hi_deleteselectedsegments()")

    def hi_detachdefault(self):
        self.callfemm("hi_detachdefault()")

    def hi_detachouterspace(self):
        self.callfemm("hi_detachouterspace()")

    def hi_drawarc(self, x1, y1, x2, y2, angle, maxseg):
        self.hi_addnode(x1, y1)
        self.hi_addnode(x2, y2)
        self.hi_addarc(x1, y1, x2, y2, angle, maxseg)

    def hi_drawline(self, x1, y1, x2, y2):
        self.hi_addnode(x1, y1)
        self.hi_addnode(x2, y2)
        self.hi_addsegment(x1, y1, x2, y2)

    def hi_drawrectangle(self, x1, y1, x2, y2):
        self.hi_drawline(x1, y1, x2, y1)
        self.hi_drawline(x2, y1, x2, y2)
        self.hi_drawline(x2, y2, x1, y2)
        self.hi_drawline(x1, y2, x1, y1)

    def hi_getmaterial(self, matname):
        self.callfemm("hi_getmaterial(" + self.quote(matname) + ")")

    def hi_hidegrid(self):
        self.callfemm("hi_hidegrid()")

    def hi_hidenames(self):
        self.callfemm("hi_shownames(0)")

    def hi_loadsolution(self):
        self.callfemm("hi_loadsolution()")

    def hi_makeABC(self, *arg):
        self.callfemm("hi_makeABC" + self.doargs(*arg))

    def hi_maximize(self):
        self.callfemm("hi_maximize()")

    def hi_minimize(self):
        self.callfemm("hi_minimize()")

    def hi_mirror(self, x1, y1, x2, y2):
        self.callfemm(
            "hi_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def hi_mirror2(self, x1, y1, x2, y2, editaction):
        self.callfemm(
            "hi_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.num(editaction)
            + ")"
        )

    def hi_modifyboundprop(self, *arg):
        self.callfemm("hi_modifyboundprop" + self.doargs(*arg))

    def hi_modifyconductorprop(self, *arg):
        self.callfemm("hi_modifyconductorprop" + self.doargs(*arg))

    def hi_modifymaterial(self, *arg):
        self.callfemm("hi_modifymaterial" + self.doargs(*arg))

    def hi_modifypointprop(self, *arg):
        self.callfemm("hi_modifypointprop" + self.doargs(*arg))

    def hi_moverotate(self, bx, by, shiftangle):
        self.callfemm(
            "hi_moverotate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(shiftangle)
            + ")"
        )

    def hi_movetranslate(self, bx, by):
        self.callfemm("hi_movetranslate(" + self.numc(bx) + self.num(by) + ")")

    def hi_movetranslate2(self, bx, by, editaction):
        self.callfemm(
            "hi_movetranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(editaction)
            + ")"
        )

    def hi_probdef(self, *arg):
        self.callfemm("hi_probdef" + self.doargs(*arg))

    def hi_purgemesh(self):
        self.callfemm("hi_purgemesh()")

    def hi_readdxf(self, docname):
        self.callfemm("print(hi_readdxf(" + self.quote(docname) + "))")

    def hi_refreshview(self):
        self.callfemm("hi_refreshview()")

    def hi_resize(self, nWidth, nHeight):
        self.callfemm("hi_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def hi_restore(self):
        self.callfemm("hi_restore()")

    def hi_saveas(self, fn):
        self.callfemm("hi_saveas(" + self.quote(self.fixpath(fn)) + ")")

    def hi_savebitmap(self, n):
        self.callfemm("hi_savebitmap(" + self.quote(n) + ")")

    def hi_savedxf(self, docname):
        self.callfemm_noeval("hi_savedxf(" + self.quote(docname) + ")")

    def hi_savemetafile(self, n):
        self.callfemm("hi_savemetafile(" + self.quote(n) + ")")

    def hi_scale(self, bx, by, sc):
        self.callfemm("hi_scale(" + self.numc(bx) + self.numc(by) + self.numc(sc) + ")")

    def hi_scale2(self, bx, by, sc, ea):
        self.callfemm(
            "hi_scale("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(sc)
            + self.num(ea)
            + ")"
        )

    def hi_selectarcsegment(self, x, y):
        return self.callfemm("hi_selectarcsegment(" + self.numc(x) + self.num(y) + ")")

    def hi_selectcircle(self, *arg):
        self.callfemm("hi_selectcircle" + self.doargs(*arg))

    def hi_selectgroup(self, gr):
        self.callfemm("hi_selectgroup(" + self.num(gr) + ")")

    def hi_selectlabel(self, x, y):
        return self.callfemm("hi_selectlabel(" + self.numc(x) + self.num(y) + ")")

    def hi_selectnode(self, x, y):
        return self.callfemm("hi_selectnode(" + self.numc(x) + self.num(y) + ")")

    def hi_selectrectangle(self, *arg):
        self.callfemm("hi_selectrectangle" + self.doargs(*arg))

    def hi_selectsegment(self, x, y):
        return self.callfemm("hi_selectsegment(" + self.numc(x) + self.num(y) + ")")

    def hi_setarcsegmentprop(self, maxsegdeg, propname, hide, group, incond):
        self.callfemm(
            "hi_setarcsegmentprop("
            + self.numc(maxsegdeg)
            + self.quotec(propname)
            + self.numc(hide)
            + self.numc(group)
            + self.quote(incond)
            + ")"
        )

    def hi_setblockprop(self, blockname, automesh, meshsize, group):
        self.callfemm(
            "hi_setblockprop("
            + self.quotec(blockname)
            + self.numc(automesh)
            + self.numc(meshsize)
            + self.num(group)
            + ")"
        )

    def hi_seteditmode(self, editmode):
        self.callfemm("hi_seteditmode(" + self.quote(editmode) + ")")

    def hi_setfocus(self, docname):
        self.callfemm("hi_setfocus(" + self.quote(docname) + ")")

    def hi_setgrid(self, density, ptype):
        self.callfemm("hi_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def hi_setgroup(self, n):
        return self.callfemm("hi_setgroup(" + self.num(n) + ")")

    def hi_setnodeprop(self, nodeprop, groupno, inconductor):
        self.callfemm(
            "hi_setnodeprop("
            + self.quotec(nodeprop)
            + self.numc(groupno)
            + self.quote(inconductor)
            + ")"
        )

    def hi_setsegmentprop(self, pn, es, am, hi, gr, inconductor):
        self.callfemm(
            "hi_setsegmentprop("
            + self.quotec(pn)
            + self.numc(es)
            + self.numc(am)
            + self.numc(hi)
            + self.numc(gr)
            + self.quote(inconductor)
            + ")"
        )

    def hi_showgrid(self):
        self.callfemm("hi_showgrid()")

    def hi_showmesh(self):
        self.callfemm("hi_showmesh()")

    def hi_shownames(self):
        self.callfemm("hi_shownames(1)")

    def hi_smartmesh(self, n):
        self.callfemm("hi_smartmesh(" + self.num(n) + ")")

    def hi_snapgridoff(self):
        self.callfemm('hi_gridsnap("off")')

    def hi_snapgridon(self):
        self.callfemm('hi_gridsnap("on")')

    def hi_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "hi_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def hi_zoomin(self):
        self.callfemm("hi_zoomin()")

    def hi_zoomnatural(self):
        self.callfemm("hi_zoomnatural()")

    def hi_zoomout(self):
        self.callfemm("hi_zoomout()")

    def hideconsole(self):
        self.callfemm("hideconsole()")

    def hidepointprops(self):
        self.callfemm("hidepointprops()")

    def ho_addcontour(self, x, y):
        self.callfemm("ho_addcontour(" + self.numc(x) + self.num(y) + ")")

    def ho_bendcontour(self, tta, dtta):
        self.callfemm("ho_bendcontour(" + self.numc(tta) + self.num(dtta) + ")")

    def ho_blockintegral(self, ptype):
        return self.callfemm("ho_blockintegral(" + self.num(ptype) + ")")

    def ho_clearblock(self):
        self.callfemm("ho_clearblock()")

    def ho_clearcontour(self):
        self.callfemm("ho_clearcontour()")

    def ho_close(self):
        self.callfemm("ho_close()")

    def ho_getconductorproperties(self, pname):
        return self.callfemm("ho_getconductorproperties(" + self.quote(pname) + ")")

    def ho_getelement(self, n):
        return self.callfemm("ho_getelement(" + self.num(n) + ")")

    def ho_getf(self, x, y):
        z = self.ho_getpointvalues(x, y)
        return [z[1], z[2]]

    def ho_getg(self, x, y):
        z = self.ho_getpointvalues(x, y)
        return [z[3], z[4]]

    def ho_getk(self, x, y):
        z = self.ho_getpointvalues(x, y)
        return [z[5], z[6]]

    def ho_getnode(self, n):
        return self.callfemm("ho_getnode(" + self.num(n) + ")")

    def ho_getpointvalues(self, x, y):
        z = self.callfemm("ho_getpointvalues(" + self.numc(x) + self.num(y) + ")")
        if len(z) > 0:
            return z
        return [0, 0, 0, 0, 0, 0, 0]

    def ho_getprobleminfo(self):
        return self.callfemm("ho_getprobleminfo()")

    def ho_gett(self, x, y):
        z = self.ho_getpointvalues(x, y)
        return z[0]

    def ho_groupselectblock(self, *arg):
        self.callfemm("ho_groupselectblock" + self.doargs(*arg))

    def ho_hidecontourplot(self):
        self.callfemm("ho_hidecontourplot()")

    def ho_hidedensityplot(self):
        self.callfemm("ho_hidedensityplot()")

    def ho_hidegrid(self):
        self.callfemm("ho_hidegrid()")

    def ho_hidemesh(self):
        self.callfemm("ho_hidemesh()")

    def ho_hidenames(self):
        self.callfemm("ho_shownames(0)")

    def ho_hidepoints(self):
        self.callfemm("ho_hidepoints()")

    def ho_lineintegral(self, ptype):
        return self.callfemm("ho_lineintegral(" + self.num(ptype) + ")")

    def ho_makeplot(self, *arg):
        self.callfemm("ho_makeplot" + self.doargs(*arg))

    def ho_maximize(self):
        self.callfemm("ho_maximize()")

    def ho_minimize(self):
        self.callfemm("ho_minimize()")

    def ho_numelements(self):
        return self.callfemm("ho_numelements()")

    def ho_numnodes(self):
        return self.callfemm("ho_numnodes()")

    def ho_refreshview(self):
        self.callfemm("ho_refreshview()")

    def ho_reload(self):
        self.callfemm("ho_reload()")

    def ho_resize(self, nWidth, nHeight):
        self.callfemm("ho_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def ho_restore(self):
        self.callfemm("ho_restore()")

    def ho_savebitmap(self, fn):
        self.callfemm("ho_savebitmap(" + self.quote(self.fixpath(fn)) + ")")

    def ho_savemetafile(self, fn):
        self.callfemm("ho_savemetafile(" + self.quote(self.fixpath(fn)) + ")")

    def ho_selectblock(self, x, y):
        self.callfemm("ho_selectblock(" + self.numc(x) + self.num(y) + ")")

    def ho_selectpoint(self, x, y):
        self.callfemm("ho_selectpoint(" + self.numc(x) + self.num(y) + ")")

    def ho_seteditmode(self, mode):
        self.callfemm("ho_seteditmode(" + self.quote(mode) + ")")

    def ho_setgrid(self, density, ptype):
        self.callfemm("ho_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def ho_showcontourplot(self, numcontours, al, au):
        self.callfemm(
            "ho_showcontourplot("
            + self.numc(numcontours)
            + self.numc(al)
            + self.num(au)
            + ")"
        )

    def ho_showdensityplot(self, legend, gscale, ptype, bu, bl):
        self.callfemm(
            "ho_showdensityplot("
            + self.numc(legend)
            + self.numc(gscale)
            + self.numc(ptype)
            + self.numc(bu)
            + self.num(bl)
            + ")"
        )

    def ho_showgrid(self):
        self.callfemm("ho_showgrid()")

    def ho_showmesh(self):
        self.callfemm("ho_showmesh()")

    def ho_shownames(self):
        self.callfemm("ho_shownames(1)")

    def ho_showpoints(self):
        self.callfemm("ho_showpoints()")

    def ho_showvectorplot(self, *arg):
        self.callfemm("ho_showvectorplot" + self.doargs(*arg))

    def ho_smooth(self, flag):
        self.callfemm("ho_smooth(" + self.quote(flag) + ")")

    def ho_smoothoff(self):
        self.callfemm('ho_smooth("off")')

    def ho_smoothon(self):
        self.callfemm('ho_smooth("on")')

    def ho_snapgrid(self, flag):
        self.callfemm("ho_gridsnap(" + self.quote(flag) + ")")

    def ho_snapgridoff(self):
        self.callfemm('ho_gridsnap("off")')

    def ho_snapgridon(self):
        self.callfemm('ho_gridsnap("on")')

    def ho_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "ho_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def ho_zoomin(self):
        self.callfemm("ho_zoomin()")

    def ho_zoomnatural(self):
        self.callfemm("ho_zoomnatural()")

    def ho_zoomout(self):
        self.callfemm("ho_zoomout()")

    def IEC(self, iec):
        return 7.959159641581004 * exp(-0.11519673572274754 * iec)

    def main_maximize(self):
        self.callfemm("main_maximize()")

    def main_minimize(self):
        self.callfemm("main_minimize()")

    def main_resize(self, nWidth, nHeight):
        self.callfemm("main_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def messagebox(self, msg):
        self.callfemm("messagebox(" + self.quote(msg) + ")")

    def mi_addarc(self, x1, y1, x2, y2, angle, maxseg):
        self.callfemm(
            "mi_addarc("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.numc(angle)
            + self.num(maxseg)
            + ")"
        )

    def mi_addbhpoint(self, name, b, h):
        self.callfemm(
            "mi_addbhpoint(" + self.quotec(name) + self.numc(b) + self.num(h) + ")"
        )

    def mi_addblocklabel(self, x, y):
        self.callfemm("mi_addblocklabel(" + self.numc(x) + self.num(y) + ")")

    def mi_addboundprop(self, *arg):
        self.callfemm("mi_addboundprop" + self.doargs(*arg))

    def mi_addcircprop(self, pname, ic, ptype):
        self.callfemm(
            "mi_addcircprop("
            + self.quotec(pname)
            + self.numc(ic)
            + self.num(ptype)
            + ")"
        )

    def mi_addmaterial(self, *arg):
        self.callfemm("mi_addmaterial" + self.doargs(*arg))

    def mi_addnode(self, x, y):
        self.callfemm("mi_addnode(" + self.numc(x) + self.num(y) + ")")

    def mi_addpointprop(self, pname, ap, jp):
        self.callfemm(
            "mi_addpointprop(" + self.quotec(pname) + self.numc(ap) + self.num(jp) + ")"
        )

    def mi_addsegment(self, x1, y1, x2, y2):
        self.callfemm(
            "mi_addsegment("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def mi_analyze(self, *arg):
        self.callfemm("mi_analyze" + self.doargs(*arg))

    def mi_analyse(self, n):
        self.mi_analyze(n)

    def mi_attachdefault(self):
        self.callfemm("mi_attachdefault()")

    def mi_attachouterspace(self):
        self.callfemm("mi_attachouterspace()")

    def mi_clearbhpoints(self, n):
        self.callfemm("mi_clearbhpoints(" + self.quote(n) + ")")

    def mi_clearselected(self):
        self.callfemm("mi_clearselected()")

    def mi_close(self):
        self.callfemm("mi_close()")

    def mi_copyrotate(self, bx, by, angle, copies):
        self.callfemm(
            "mi_copyrotate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(angle)
            + self.num(copies)
            + ")"
        )

    def mi_copyrotate2(self, bx, by, angle, copies, editaction):
        self.callfemm(
            "mi_copyrotate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(angle)
            + self.numc(copies)
            + self.num(editaction)
            + ")"
        )

    def mi_copytranslate(self, bx, by, copies):
        self.callfemm(
            "mi_copytranslate(" + self.numc(bx) + self.numc(by) + self.num(copies) + ")"
        )

    def mi_copytranslate2(self, bx, by, copies, editaction):
        self.callfemm(
            "mi_copytranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(copies)
            + self.num(editaction)
            + ")"
        )

    def mi_createmesh(self):
        return self.callfemm("mi_createmesh()")

    def mi_createradius(self, x, y, r):
        self.callfemm(
            "mi_createradius(" + self.numc(x) + self.numc(y) + self.num(r) + ")"
        )

    def mi_defineouterspace(self, Zo, Ro, Ri):
        self.callfemm(
            "mi_defineouterspace(" + self.numc(Zo) + self.numc(Ro) + self.num(Ri) + ")"
        )

    def mi_deleteboundprop(self, n):
        self.callfemm("mi_deleteboundprop(" + self.quote(n) + ")")

    def mi_deletecircuit(self, n):
        self.callfemm("mi_deletecircuit(" + self.quote(n) + ")")

    def mi_deletematerial(self, n):
        self.callfemm("mi_deletematerial(" + self.quote(n) + ")")

    def mi_deletepointprop(self, n):
        self.callfemm("mi_deletepointprop(" + self.quote(n) + ")")

    def mi_deleteselected(self):
        self.callfemm("mi_deleteselected()")

    def mi_deleteselectedarcsegments(self):
        self.callfemm("mi_deleteselectedarcsegments()")

    def mi_deleteselectedlabels(self):
        self.callfemm("mi_deleteselectedlabels()")

    def mi_deleteselectednodes(self):
        self.callfemm("mi_deleteselectednodes()")

    def mi_deleteselectedsegments(self):
        self.callfemm("mi_deleteselectedsegments()")

    def mi_detachdefault(self):
        self.callfemm("mi_detachdefault()")

    def mi_detachouterspace(self):
        self.callfemm("mi_detachouterspace()")

    def mi_drawarc(self, x1, y1, x2, y2, angle, maxseg):
        self.mi_addnode(x1, y1)
        self.mi_addnode(x2, y2)
        self.mi_addarc(x1, y1, x2, y2, angle, maxseg)

    def mi_drawline(self, x1, y1, x2, y2):
        self.mi_addnode(x1, y1)
        self.mi_addnode(x2, y2)
        self.mi_addsegment(x1, y1, x2, y2)

    def mi_drawrectangle(self, x1, y1, x2, y2):
        self.mi_drawline(x1, y1, x2, y1)
        self.mi_drawline(x2, y1, x2, y2)
        self.mi_drawline(x2, y2, x1, y2)
        self.mi_drawline(x1, y2, x1, y1)

    def mi_getmaterial(self, matname):
        self.callfemm("mi_getmaterial(" + self.quote(matname) + ")")

    def mi_hidegrid(self):
        self.callfemm("mi_hidegrid()")

    def mi_hidenames(self):
        self.callfemm("mi_shownames(0)")

    def mi_loadsolution(self):
        self.callfemm("mi_loadsolution()")

    def mi_makeABC(self, *arg):
        self.callfemm("mi_makeABC" + self.doargs(*arg))

    def mi_maximize(self):
        self.callfemm("mi_maximize()")

    def mi_minimize(self):
        self.callfemm("mi_minimize()")

    def mi_mirror(self, x1, y1, x2, y2):
        self.callfemm(
            "mi_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def mi_mirror2(self, x1, y1, x2, y2, editaction):
        self.callfemm(
            "mi_mirror("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.numc(y2)
            + self.num(editaction)
            + ")"
        )

    def mi_modifyboundprop(self, *arg):
        self.callfemm("mi_modifyboundprop" + self.doargs(*arg))

    def mi_modifycircprop(self, *arg):
        self.callfemm("mi_modifycircprop" + self.doargs(*arg))

    def mi_modifymaterial(self, *arg):
        self.callfemm("mi_modifymaterial" + self.doargs(*arg))

    def mi_modifypointprop(self, *arg):
        self.callfemm("mi_modifypointprop" + self.doargs(*arg))

    def mi_moverotate(self, bx, by, shiftangle):
        self.callfemm(
            "mi_moverotate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(shiftangle)
            + ")"
        )

    def mi_movetranslate(self, bx, by):
        self.callfemm("mi_movetranslate(" + self.numc(bx) + self.num(by) + ")")

    def mi_movetranslate2(self, bx, by, editaction):
        self.callfemm(
            "mi_movetranslate("
            + self.numc(bx)
            + self.numc(by)
            + self.num(editaction)
            + ")"
        )

    def mi_probdef(self, *arg):
        self.callfemm("mi_probdef" + self.doargs(*arg))

    def mi_purgemesh(self):
        self.callfemm("mi_purgemesh()")

    def mi_readdxf(self, docname):
        self.callfemm("print(mi_readdxf(" + self.quote(docname) + "))")

    def mi_refreshview(self):
        self.callfemm("mi_refreshview()")

    def mi_resize(self, nWidth, nHeight):
        self.callfemm("mi_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def mi_restore(self):
        self.callfemm("mi_restore()")

    def mi_saveas(self, fn):
        self.callfemm("mi_saveas(" + self.quote(self.fixpath(fn)) + ")")

    def mi_savebitmap(self, n):
        self.callfemm("mi_savebitmap(" + self.quote(n) + ")")

    def mi_savedxf(self, docname):
        self.callfemm_noeval("mi_savedxf(" + self.quote(docname) + ")")

    def mi_savemetafile(self, n):
        self.callfemm("mi_savemetafile(" + self.quote(n) + ")")

    def mi_scale(self, bx, by, sc):
        self.callfemm("mi_scale(" + self.numc(bx) + self.numc(by) + self.numc(sc) + ")")

    def mi_scale2(self, bx, by, sc, ea):
        self.callfemm(
            "mi_scale("
            + self.numc(bx)
            + self.numc(by)
            + self.numc(sc)
            + self.num(ea)
            + ")"
        )

    def mi_selectarcsegment(self, x, y):
        return self.callfemm("mi_selectarcsegment(" + self.numc(x) + self.num(y) + ")")

    def mi_selectcircle(self, *arg):
        self.callfemm("mi_selectcircle" + self.doargs(*arg))

    def mi_selectgroup(self, gr):
        self.callfemm("mi_selectgroup(" + self.num(gr) + ")")

    def mi_selectlabel(self, x, y):
        return self.callfemm("mi_selectlabel(" + self.numc(x) + self.num(y) + ")")

    def mi_selectnode(self, x, y):
        return self.callfemm("mi_selectnode(" + self.numc(x) + self.num(y) + ")")

    def mi_selectrectangle(self, *arg):
        self.callfemm("mi_selectrectangle" + self.doargs(*arg))

    def mi_selectsegment(self, x, y):
        return self.callfemm("mi_selectsegment(" + self.numc(x) + self.num(y) + ")")

    def mi_setarcsegmentprop(self, maxsegdeg, propname, hide, group):
        self.callfemm(
            "mi_setarcsegmentprop("
            + self.numc(maxsegdeg)
            + self.quotec(propname)
            + self.numc(hide)
            + self.num(group)
            + ")"
        )

    def mi_setblockprop(self, *arg):
        self.callfemm("mi_setblockprop" + self.doargs(*arg))

    def mi_setcurrent(self, name, x):
        self.mi_modifycircprop(name, 1, x)

    def mi_seteditmode(self, editmode):
        self.callfemm("mi_seteditmode(" + self.quote(editmode) + ")")

    def mi_setfocus(self, docname):
        self.callfemm("mi_setfocus(" + self.quote(docname) + ")")

    def mi_setgrid(self, density, ptype):
        self.callfemm("mi_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def mi_setgroup(self, n):
        return self.callfemm("mi_setgroup(" + self.num(n) + ")")

    def mi_setnodeprop(self, nodeprop, groupno):
        self.callfemm(
            "mi_setnodeprop(" + self.quotec(nodeprop) + self.num(groupno) + ")"
        )

    def mi_setprevious(self, *arg):
        self.callfemm("mi_setprevious" + self.doargs(*arg))

    def mi_setsegmentprop(self, pn, es, am, hi, gr):
        self.callfemm(
            "mi_setsegmentprop("
            + self.quotec(pn)
            + self.numc(es)
            + self.numc(am)
            + self.numc(hi)
            + self.num(gr)
            + ")"
        )

    def mi_showgrid(self):
        self.callfemm("mi_showgrid()")

    def mi_showmesh(self):
        self.callfemm("mi_showmesh()")

    def mi_shownames(self):
        self.callfemm("mi_shownames(1)")

    def mi_smartmesh(self, n):
        self.callfemm("mi_smartmesh(" + self.num(n) + ")")

    def mi_snapgridoff(self):
        self.callfemm('mi_gridsnap("off")')

    def mi_snapgridon(self):
        self.callfemm('mi_gridsnap("on")')

    def mi_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "mi_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def mi_zoomin(self):
        self.callfemm("mi_zoomin()")

    def mi_zoomnatural(self):
        self.callfemm("mi_zoomnatural()")

    def mi_zoomout(self):
        self.callfemm("mi_zoomout()")

    def mo_addcontour(self, x, y):
        self.callfemm("mo_addcontour(" + self.numc(x) + self.num(y) + ")")

    def mo_bendcontour(self, tta, dtta):
        self.callfemm("mo_bendcontour(" + self.numc(tta) + self.num(dtta) + ")")

    def mo_blockintegral(self, ptype):
        return self.callfemm("mo_blockintegral(" + self.num(ptype) + ")")

    def mo_clearblock(self):
        self.callfemm("mo_clearblock()")

    def mo_clearcontour(self):
        self.callfemm("mo_clearcontour()")

    def mo_close(self):
        self.callfemm("mo_close()")

    def mo_gapintegral(self, bdryname, inttype):
        return self.callfemm(
            "mo_gapintegral(" + self.quotec(bdryname) + self.num(inttype) + ")"
        )

    def mo_geta(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[0]

    def mo_getb(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return [z[1], z[2]]

    def mo_getcircuitproperties(self, name):
        return self.callfemm("mo_getcircuitproperties(" + self.quote(name) + ")")

    def mo_getconductivity(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[3]

    def mo_getelement(self, n):
        return self.callfemm("mo_getelement(" + self.num(n) + ")")

    def mo_getenergydensity(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[4]

    def mo_getfill(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[13]

    def mo_getgapa(self, bdryname, angle):
        return self.callfemm(
            "mo_getgapa(" + self.quotec(bdryname) + self.num(angle) + ")"
        )

    def mo_getgapb(self, bdryname, angle):
        return self.callfemm(
            "mo_getgapb(" + self.quotec(bdryname) + self.num(angle) + ")"
        )

    def mo_getgapharmonics(self, *arg):
        return self.callfemm("mo_getgapharmonics" + self.doargs(*arg))

    def mo_geth(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return [z[5], z[6]]

    def mo_getj(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[7] + z[8]

    def mo_getmu(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return [z[9], z[10]]

    def mo_getnode(self, n):
        return self.callfemm("mo_getnode(" + self.num(n) + ")")

    def mo_getpe(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[11]

    def mo_getph(self, x, y):
        z = self.mo_getpointvalues(x, y)
        return z[12]

    def mo_getpointvalues(self, x, y):
        z = self.callfemm("mo_getpointvalues(" + self.numc(x) + self.num(y) + ")")
        if len(z) > 0:
            return z
        return [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    def mo_getprobleminfo(self):
        return self.callfemm("mo_getprobleminfo()")

    def mo_groupselectblock(self, *arg):
        self.callfemm("mo_groupselectblock" + self.doargs(*arg))

    def mo_hidecontourplot(self):
        self.callfemm("mo_hidecontourplot()")

    def mo_hidedensityplot(self):
        self.callfemm("mo_hidedensityplot()")

    def mo_hidegrid(self):
        self.callfemm("mo_hidegrid()")

    def mo_hidemesh(self):
        self.callfemm("mo_hidemesh()")

    def mo_hidenames(self):
        self.callfemm("mo_shownames(0)")

    def mo_hidepoints(self):
        self.callfemm("mo_hidepoints()")

    def mo_lineintegral(self, ptype):
        return self.callfemm("mo_lineintegral(" + self.num(ptype) + ")")

    def mo_makeplot(self, *arg):
        self.callfemm("mo_makeplot" + self.doargs(*arg))

    def mo_maximize(self):
        self.callfemm("mo_maximize()")

    def mo_minimize(self):
        self.callfemm("mo_minimize()")

    def mo_numelements(self):
        return self.callfemm("mo_numelements()")

    def mo_numnodes(self):
        return self.callfemm("mo_numnodes()")

    def mo_refreshview(self):
        self.callfemm("mo_refreshview()")

    def mo_reload(self):
        self.callfemm("mo_reload()")

    def mo_resize(self, nWidth, nHeight):
        self.callfemm("mo_resize(" + self.numc(nWidth) + self.num(nHeight) + ")")

    def mo_restore(self):
        self.callfemm("mo_restore()")

    def mo_savebitmap(self, fn):
        self.callfemm("mo_savebitmap(" + self.quote(self.fixpath(fn)) + ")")

    def mo_savemetafile(self, fn):
        self.callfemm("mo_savemetafile(" + self.quote(self.fixpath(fn)) + ")")

    def mo_selectblock(self, x, y):
        self.callfemm("mo_selectblock(" + self.numc(x) + self.num(y) + ")")

    def mo_selectpoint(self, x, y):
        self.callfemm("mo_selectpoint(" + self.numc(x) + self.num(y) + ")")

    def mo_seteditmode(self, mode):
        self.callfemm("mo_seteditmode(" + self.quote(mode) + ")")

    def mo_setgrid(self, density, ptype):
        self.callfemm("mo_setgrid(" + self.numc(density) + self.quote(ptype) + ")")

    def mo_setweightingscheme(self, n):
        self.callfemm("mo_setweightingscheme(" + self.num(n) + ")")

    def mo_showcontourplot(self, numcontours, al, au, ptype):
        self.callfemm(
            "mo_showcontourplot("
            + self.numc(numcontours)
            + self.numc(al)
            + self.numc(au)
            + self.quote(ptype)
            + ")"
        )

    def mo_showdensityplot(self, legend, gscale, bu, bl, ptype):
        self.callfemm(
            "mo_showdensityplot("
            + self.numc(legend)
            + self.numc(gscale)
            + self.numc(bu)
            + self.numc(bl)
            + self.quote(ptype)
            + ")"
        )

    def mo_showgrid(self):
        self.callfemm("mo_showgrid()")

    def mo_showmesh(self):
        self.callfemm("mo_showmesh()")

    def mo_shownames(self):
        self.callfemm("mo_shownames(1)")

    def mo_showpoints(self):
        self.callfemm("mo_showpoints()")

    def mo_showvectorplot(self, *arg):
        self.callfemm("mo_showvectorplot" + self.doargs(*arg))

    def mo_smooth(self, flag):
        self.callfemm("mo_smooth(" + self.quote(flag) + ")")

    def mo_smoothoff(self):
        self.callfemm('mo_smooth("off")')

    def mo_smoothon(self):
        self.callfemm('mo_smooth("on")')

    def mo_snapgrid(self, flag):
        self.callfemm("mo_gridsnap(" + self.quote(flag) + ")")

    def mo_snapgridoff(self):
        self.callfemm('mo_gridsnap("off")')

    def mo_snapgridon(self):
        self.callfemm('mo_gridsnap("on")')

    def mo_zoom(self, x1, y1, x2, y2):
        self.callfemm(
            "mo_zoom("
            + self.numc(x1)
            + self.numc(y1)
            + self.numc(x2)
            + self.num(y2)
            + ")"
        )

    def mo_zoomin(self):
        self.callfemm("mo_zoomin()")

    def mo_zoomnatural(self):
        self.callfemm("mo_zoomnatural()")

    def mo_zoomout(self):
        self.callfemm("mo_zoomout()")

    def opendocument(self, fn):
        self.callfemm("open(" + self.quote(self.fixpath(fn)) + ")")

    def prompt(self, msg):
        return self.callfemm("prompt(" + self.quote(msg) + ")")

    def showconsole(self):
        self.callfemm("showconsole()")

    def showpointprops(self):
        self.callfemm("showpointprops()")

    def smartmesh(self, n):
        self.callfemm("smartmesh(" + self.num(n) + ")")



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/Machine/LamSquirrelCage.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/pyleecan/tree/master/pyleecan/Methods/Machine/LamSquirrelCage
"""

from os import linesep
from sys import getsizeof
from logging import getLogger
from ._check import check_var, raise_
from ..Functions.get_logger import get_logger
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from .LamSlotWind import LamSlotWind

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.Machine.LamSquirrelCage.build_geometry import build_geometry
except ImportError as error:
    build_geometry = error

try:
    from ..Methods.Machine.LamSquirrelCage.check import check
except ImportError as error:
    check = error

try:
    from ..Methods.Machine.LamSquirrelCage.comp_length_ring import comp_length_ring
except ImportError as error:
    comp_length_ring = error

try:
    from ..Methods.Machine.LamSquirrelCage.plot import plot
except ImportError as error:
    plot = error

try:
    from ..Methods.Machine.LamSquirrelCage.comp_number_phase_eq import (
        comp_number_phase_eq,
    )
except ImportError as error:
    comp_number_phase_eq = error

try:
    from ..Methods.Machine.LamSquirrelCage.comp_periodicity import comp_periodicity
except ImportError as error:
    comp_periodicity = error

try:
    from ..Methods.Machine.LamSquirrelCage.comp_surface_ring import comp_surface_ring
except ImportError as error:
    comp_surface_ring = error

try:
    from ..Methods.Machine.LamSquirrelCage.comp_resistance_wind import (
        comp_resistance_wind,
    )
except ImportError as error:
    comp_resistance_wind = error


from ._check import InitUnKnowClassError
from .Material import Material
from .Winding import Winding
from .Slot import Slot
from .Hole import Hole
from .Notch import Notch


class LamSquirrelCage(LamSlotWind):
    """squirrel cages lamination"""

    VERSION = 1

    # Check ImportError to remove unnecessary dependencies in unused method
    # cf Methods.Machine.LamSquirrelCage.build_geometry
    if isinstance(build_geometry, ImportError):
        build_geometry = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use LamSquirrelCage method build_geometry: "
                    + str(build_geometry)
                )
            )
        )
    else:
        build_geometry = build_geometry
    # cf Methods.Machine.LamSquirrelCage.check
    if isinstance(check, ImportError):
        check = property(
            fget=lambda x: raise_(
                ImportError("Can't use LamSquirrelCage method check: " + str(check))
            )
        )
    else:
        check = check
    # cf Methods.Machine.LamSquirrelCage.comp_length_ring
    if isinstance(comp_length_ring, ImportError):
        comp_length_ring = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use LamSquirrelCage method comp_length_ring: "
                    + str(comp_length_ring)
                )
            )
        )
    else:
        comp_length_ring = comp_length_ring
    # cf Methods.Machine.LamSquirrelCage.plot
    if isinstance(plot, ImportError):
        plot = property(
            fget=lambda x: raise_(
                ImportError("Can't use LamSquirrelCage method plot: " + str(plot))
            )
        )
    else:
        plot = plot
    # cf Methods.Machine.LamSquirrelCage.comp_number_phase_eq
    if isinstance(comp_number_phase_eq, ImportError):
        comp_number_phase_eq = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use LamSquirrelCage method comp_number_phase_eq: "
                    + str(comp_number_phase_eq)
                )
            )
        )
    else:
        comp_number_phase_eq = comp_number_phase_eq
    # cf Methods.Machine.LamSquirrelCage.comp_periodicity
    if isinstance(comp_periodicity, ImportError):
        comp_periodicity = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use LamSquirrelCage method comp_periodicity: "
                    + str(comp_periodicity)
                )
            )
        )
    else:
        comp_periodicity = comp_periodicity
    # cf Methods.Machine.LamSquirrelCage.comp_surface_ring
    if isinstance(comp_surface_ring, ImportError):
        comp_surface_ring = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use LamSquirrelCage method comp_surface_ring: "
                    + str(comp_surface_ring)
                )
            )
        )
    else:
        comp_surface_ring = comp_surface_ring
    # cf Methods.Machine.LamSquirrelCage.comp_resistance_wind
    if isinstance(comp_resistance_wind, ImportError):
        comp_resistance_wind = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use LamSquirrelCage method comp_resistance_wind: "
                    + str(comp_resistance_wind)
                )
            )
        )
    else:
        comp_resistance_wind = comp_resistance_wind
    # save and copy methods are available in all object
    save = save
    copy = copy
    # get_logger method is available in all object
    get_logger = get_logger

    def __init__(
        self,
        Hscr=0.03,
        Lscr=0.015,
        ring_mat=-1,
        Ksfill=None,
        winding=-1,
        slot=-1,
        L1=0.35,
        mat_type=-1,
        Nrvd=0,
        Wrvd=0,
        Kf1=0.95,
        is_internal=True,
        Rint=0,
        Rext=1,
        is_stator=True,
        axial_vent=-1,
        notch=-1,
        init_dict=None,
        init_str=None,
    ):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for pyleecan type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for pyleecan Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "Hscr" in list(init_dict.keys()):
                Hscr = init_dict["Hscr"]
            if "Lscr" in list(init_dict.keys()):
                Lscr = init_dict["Lscr"]
            if "ring_mat" in list(init_dict.keys()):
                ring_mat = init_dict["ring_mat"]
            if "Ksfill" in list(init_dict.keys()):
                Ksfill = init_dict["Ksfill"]
            if "winding" in list(init_dict.keys()):
                winding = init_dict["winding"]
            if "slot" in list(init_dict.keys()):
                slot = init_dict["slot"]
            if "L1" in list(init_dict.keys()):
                L1 = init_dict["L1"]
            if "mat_type" in list(init_dict.keys()):
                mat_type = init_dict["mat_type"]
            if "Nrvd" in list(init_dict.keys()):
                Nrvd = init_dict["Nrvd"]
            if "Wrvd" in list(init_dict.keys()):
                Wrvd = init_dict["Wrvd"]
            if "Kf1" in list(init_dict.keys()):
                Kf1 = init_dict["Kf1"]
            if "is_internal" in list(init_dict.keys()):
                is_internal = init_dict["is_internal"]
            if "Rint" in list(init_dict.keys()):
                Rint = init_dict["Rint"]
            if "Rext" in list(init_dict.keys()):
                Rext = init_dict["Rext"]
            if "is_stator" in list(init_dict.keys()):
                is_stator = init_dict["is_stator"]
            if "axial_vent" in list(init_dict.keys()):
                axial_vent = init_dict["axial_vent"]
            if "notch" in list(init_dict.keys()):
                notch = init_dict["notch"]
        # Set the properties (value check and convertion are done in setter)
        self.Hscr = Hscr
        self.Lscr = Lscr
        self.ring_mat = ring_mat
        # Call LamSlotWind init
        super(LamSquirrelCage, self).__init__(
            Ksfill=Ksfill,
            winding=winding,
            slot=slot,
            L1=L1,
            mat_type=mat_type,
            Nrvd=Nrvd,
            Wrvd=Wrvd,
            Kf1=Kf1,
            is_internal=is_internal,
            Rint=Rint,
            Rext=Rext,
            is_stator=is_stator,
            axial_vent=axial_vent,
            notch=notch,
        )
        # The class is frozen (in LamSlotWind init), for now it's impossible to
        # add new properties

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        LamSquirrelCage_str = ""
        # Get the properties inherited from LamSlotWind
        LamSquirrelCage_str += super(LamSquirrelCage, self).__str__()
        LamSquirrelCage_str += "Hscr = " + str(self.Hscr) + linesep
        LamSquirrelCage_str += "Lscr = " + str(self.Lscr) + linesep
        if self.ring_mat is not None:
            tmp = self.ring_mat.__str__().replace(linesep, linesep + "\t").rstrip("\t")
            LamSquirrelCage_str += "ring_mat = " + tmp
        else:
            LamSquirrelCage_str += "ring_mat = None" + linesep + linesep
        return LamSquirrelCage_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False

        # Check the properties inherited from LamSlotWind
        if not super(LamSquirrelCage, self).__eq__(other):
            return False
        if other.Hscr != self.Hscr:
            return False
        if other.Lscr != self.Lscr:
            return False
        if other.ring_mat != self.ring_mat:
            return False
        return True

    def compare(self, other, name="self"):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()

        # Check the properties inherited from LamSlotWind
        diff_list.extend(super(LamSquirrelCage, self).compare(other, name=name))
        if other._Hscr != self._Hscr:
            diff_list.append(name + ".Hscr")
        if other._Lscr != self._Lscr:
            diff_list.append(name + ".Lscr")
        if (other.ring_mat is None and self.ring_mat is not None) or (
            other.ring_mat is not None and self.ring_mat is None
        ):
            diff_list.append(name + ".ring_mat None mismatch")
        elif self.ring_mat is not None:
            diff_list.extend(
                self.ring_mat.compare(other.ring_mat, name=name + ".ring_mat")
            )
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object

        # Get size of the properties inherited from LamSlotWind
        S += super(LamSquirrelCage, self).__sizeof__()
        S += getsizeof(self.Hscr)
        S += getsizeof(self.Lscr)
        S += getsizeof(self.ring_mat)
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        # Get the properties inherited from LamSlotWind
        LamSquirrelCage_dict = super(LamSquirrelCage, self).as_dict(**kwargs)
        LamSquirrelCage_dict["Hscr"] = self.Hscr
        LamSquirrelCage_dict["Lscr"] = self.Lscr
        if self.ring_mat is None:
            LamSquirrelCage_dict["ring_mat"] = None
        else:
            LamSquirrelCage_dict["ring_mat"] = self.ring_mat.as_dict(**kwargs)
        # The class name is added to the dict for deserialisation purpose
        # Overwrite the mother class name
        LamSquirrelCage_dict["__class__"] = "LamSquirrelCage"
        return LamSquirrelCage_dict

    def _set_None(self):
        """Set all the properties to None (except pyleecan object)"""

        self.Hscr = None
        self.Lscr = None
        if self.ring_mat is not None:
            self.ring_mat._set_None()
        # Set to None the properties inherited from LamSlotWind
        super(LamSquirrelCage, self)._set_None()

    def _get_Hscr(self):
        """getter of Hscr"""
        return self._Hscr

    def _set_Hscr(self, value):
        """setter of Hscr"""
        check_var("Hscr", value, "float", Vmin=0)
        self._Hscr = value

    Hscr = property(
        fget=_get_Hscr,
        fset=_set_Hscr,
        doc=u"""short circuit ring section radial height [m]

        :Type: float
        :min: 0
        """,
    )

    def _get_Lscr(self):
        """getter of Lscr"""
        return self._Lscr

    def _set_Lscr(self, value):
        """setter of Lscr"""
        check_var("Lscr", value, "float", Vmin=0)
        self._Lscr = value

    Lscr = property(
        fget=_get_Lscr,
        fset=_set_Lscr,
        doc=u"""short circuit ring section axial length

        :Type: float
        :min: 0
        """,
    )

    def _get_ring_mat(self):
        """getter of ring_mat"""
        return self._ring_mat

    def _set_ring_mat(self, value):
        """setter of ring_mat"""
        if isinstance(value, str):  # Load from file
            value = load_init_dict(value)[1]
        if isinstance(value, dict) and "__class__" in value:
            class_obj = import_class(
                "pyleecan.Classes", value.get("__class__"), "ring_mat"
            )
            value = class_obj(init_dict=value)
        elif type(value) is int and value == -1:  # Default constructor
            value = Material()
        check_var("ring_mat", value, "Material")
        self._ring_mat = value

        if self._ring_mat is not None:
            self._ring_mat.parent = self

    ring_mat = property(
        fget=_get_ring_mat,
        fset=_set_ring_mat,
        doc=u"""Material of the Rotor short circuit ring

        :Type: Material
        """,
    )



# -*- coding: utf-8 -*-

import numpy as np


def is_exist(self, new_coord, dim=2):
    """Check the existence of a node defined by its coordinates

    Parameters
    ----------
    self : PointMat
        an PointMat object
    coord : ndarray
        coordinate of the node

    Returns
    -------
        bool
            True if the element already exist
    """
    # Check the existence of the element
    delta = self.delta
    nb_node = self.nb_pt
    coords = self.coordinate

    if nb_node == 0:
        return False
    if nb_node == 1:
        dist = np.sqrt(np.sum(np.square(coords - new_coord)))
        if dist > delta:
            return False
        else:
            return True
    else:
        point = np.tile(new_coord, (nb_node, 1))
        if len(new_coord) == 3:
            dist_node = np.reshape(
                np.sqrt(
                    np.square(coords[:, 0] - point[:, 0])
                    + np.square(coords[:, 1] - point[:, 1])
                    + np.square(coords[:, 2] - point[:, 2])
                ),
                (nb_node, 1),
            )
        if len(new_coord) == 2:
            dist_node = np.reshape(
                np.sqrt(
                    np.square(coords[:, 0] - point[:, 0])
                    + np.square(coords[:, 1] - point[:, 1])
                ),
                (nb_node, 1),
            )

        min_node_list = np.argsort(dist_node, axis=0)

        if dist_node[min_node_list[0]] > delta:
            return False
        else:
            return True



# -*- coding: utf-8 -*-
from ....Functions.Winding.comp_wind_periodicity import comp_wind_periodicity
from ....Classes.Winding import Winding


def comp_periodicity(self):
    """Compute the periodicity factor of the lamination

    Parameters
    ----------
    self : LamSlotWind
        A LamSlotWind object

    Returns
    -------
    per_a : int
        Number of spatial periodicities of the lamination
    is_antiper_a : bool
        True if an spatial anti-periodicity is possible after the periodicities
    per_t : int
        Number of time periodicities of the lamination
    is_antiper_t : bool
        True if an time anti-periodicity is possible after the periodicities
    """

    if self.winding is not None and type(self.winding) is not Winding:
        sym_a, is_antisym_a = comp_wind_periodicity(self.winding.comp_connection_mat())
    else:
        sym_a, is_antisym_a = 1, False

    if is_antisym_a:
        sym_a /= 2

    return (int(sym_a), is_antisym_a, sym_a, is_antisym_a)



from ....Classes.Segment import Segment
from ....Classes.SurfLine import SurfLine
from ....Classes.Arc1 import Arc1


def build_geometry(self, alpha=0, delta=0, is_simplified=False):
    """Compute the curve (Segment) needed to plot the Slot.
    The ending point of a curve is the starting point of the next curve in
    the list

    Parameters
    ----------
    self : HoleM50
        A HoleM50 object
    alpha : float
        Angle to rotate the slot (Default value = 0) [rad]
    delta : complex
        Complex to translate the slot (Default value = 0)
    is_simplified : bool
       True to avoid line superposition

    Returns
    -------
    surf_list: list
        List of SurfLine needed to draw the HoleM50

    """

    # Get correct label for surfaces
    if self.get_is_stator():
        st = "_Stator"
    else:
        st = "_Rotor"

    # Get all the points
    point_dict = self._comp_point_coordinate()
    Z1 = point_dict["Z1"]
    Z2 = point_dict["Z2"]
    Z3 = point_dict["Z3"]
    Z4 = point_dict["Z4"]
    Z5 = point_dict["Z5"]
    Z6 = point_dict["Z6"]
    Z7 = point_dict["Z7"]
    Z8 = point_dict["Z8"]
    Z9 = point_dict["Z9"]
    Z10 = point_dict["Z10"]
    Z11 = point_dict["Z11"]
    Z8c = point_dict["Z8c"]
    Z8b = point_dict["Z8b"]
    Z1s = point_dict["Z1s"]
    Z2s = point_dict["Z2s"]
    Z3s = point_dict["Z3s"]
    Z4s = point_dict["Z4s"]
    Z5s = point_dict["Z5s"]
    Z6s = point_dict["Z6s"]
    Z7s = point_dict["Z7s"]
    Z8s = point_dict["Z8s"]
    Z9s = point_dict["Z9s"]
    Z10s = point_dict["Z10s"]
    Z11s = point_dict["Z11s"]
    Z8cs = point_dict["Z8cs"]
    Z8bs = point_dict["Z8bs"]
    Rext = self.get_Rext()

    # Create all the lines
    L_1_2 = Segment(Z1, Z2, label="")
    L_1s_2s = Segment(Z1s, Z2s, label="")
    L_1s_11s = Segment(Z1s, Z11s, label="")
    L_2_3 = Segment(Z2, Z3, label="")
    L_2s_1s = Segment(Z2s, Z1s, label="")
    L_2s_3s = Segment(Z2s, Z3s, label="")
    L_3_4 = Segment(Z3, Z4, label="Rotor_Magnet_0_Right")
    L_3s_2s = Segment(Z3s, Z2s, label="")
    L_3s_4s = Segment(Z3s, Z4s, label="Rotor_Magnet_1_Left")
    L_3_8c = Segment(Z3, Z8c, label="Rotor_Hole_0_Right")
    L_3s_8cs = Segment(Z3s, Z8cs, label="Rotor_Hole_1_Left")
    L_4_5 = Segment(Z4, Z5, label="Rotor_Magnet_0_Bottom")
    L_4s_3s = Segment(Z4s, Z3s, label="")
    L_4s_5s = Segment(Z4s, Z5s, label="Rotor_Magnet_1_Bottom")
    L_5_6 = Segment(Z5, Z6, label="Rotor_Magnet_0_Left")
    L_5s_4s = Segment(Z5s, Z4s, label="")
    L_5s_6s = Segment(Z5s, Z6s, label="Rotor_Magnet_1_Right")
    L_6_7 = Segment(Z6, Z7, label="")
    L_6s_5s = Segment(Z6s, Z5s, label="")
    L_6s_7s = Segment(Z6s, Z7s, label="")
    L_6_8b = Segment(Z6, Z8b, label="")
    L_6s_8bs = Segment(Z6s, Z8bs, label="")
    L_7_6s = Segment(Z7, Z6s, label="")
    L_7_8 = Segment(Z7, Z8, label="Rotor_Radial_Bridge")
    L_7s_6 = Segment(Z7s, Z6, label="")
    L_7s_8s = Segment(Z7s, Z8s, label="Rotor_Radial_Bridge")

    L_8_8b = Segment(Z8, Z8b, label="")
    L_8_9 = Segment(Z8, Z9, label="")
    L_8s_8bs = Segment(Z8s, Z8bs, label="Rotor_Hole_1_Top")
    L_8s_8b = Segment(Z8s, Z8b, label="Rotor_Hole_0_Top")
    L_8s_9 = Segment(Z8s, Z9, label="")
    L_8s_9s = Segment(Z8s, Z9s, label="")

    L_8b_6 = Segment(Z8b, Z6, label="Rotor_Hole_0_Left")
    L_8bs_6s = Segment(Z8bs, Z6s, label="Rotor_Hole_1_Right")
    L_8b_8c = Segment(Z8b, Z8c, label="Rotor_Magnet_0_Top")
    L_8b_8s = Segment(Z8b, Z8s, label="")
    L_8bs_8 = Segment(Z8bs, Z8, label="")
    L_8bs_8s = Segment(Z8bs, Z8s, label="Rotor_Hole_1_Top")
    L_8bs_8cs = Segment(Z8bs, Z8cs, label="Rotor_Magnet_1_Top")

    L_8c_3 = Segment(Z8c, Z3, label="")
    L_8cs_3s = Segment(Z8cs, Z3s, label="")
    L_8c_9 = Segment(Z8c, Z9, label="Rotor_Hole_0_Top")
    L_8cs_9s = Segment(Z8cs, Z9s, label="Rotor_Hole_1_Top")

    L_9_10 = Segment(Z9, Z10, label="")
    L_9s_8s = Segment(Z9s, Z8s, label="")
    L_9s_10s = Segment(Z9s, Z10s, label="")
    L_10_11 = Arc1(
        Z10,
        Z11,
        -Rext + self.H1,
        is_trigo_direction=False,
        label="Rotor_Tangential_Bridge",
    )
    L_10s_9s = Segment(Z10s, Z9s, label="")
    L_10s_11s = Arc1(
        Z10s,
        Z11s,
        Rext - self.H1,
        is_trigo_direction=True,
        label="Rotor_Tangential_Bridge",
    )
    L_11_1 = Segment(Z11, Z1, label="")
    L_11s_1s = Segment(Z11s, Z1s, label="")
    L_11s_10s = Arc1(Z11s, Z10s, -Rext + self.H1, is_trigo_direction=False)

    # Create all the surfaces for all the cases
    surf_list = list()
    # Air surface (W3) with magnet_0
    curve_list = [L_1_2, L_2_3, L_3_8c, L_8c_9]
    if self.H4 > 0:
        curve_list.append(L_9_10)
    curve_list.append(L_10_11)
    if self.H4 > 0:
        curve_list.append(L_11_1)
    point_ref = (Z1 + Z2 + Z3 + Z8c + Z9 + Z10 + Z11) / 7
    S1 = SurfLine(line_list=curve_list, label="Rotor_Hole" + st, point_ref=point_ref)

    # Magnet_0 surface
    curve_list = list()
    if is_simplified:
        curve_list.append(L_8c_3)
        curve_list.append(L_6_8b)
    else:
        if Z3 != Z4:  # Z3 == Z4 if H2 = 0
            curve_list.append(L_3_4)
        curve_list.append(L_4_5)
        if Z5 != Z6:  # Z5 == Z6 if H2 = 0
            curve_list.append(L_5_6)
        curve_list.append(L_6_8b)
        curve_list.append(L_8b_8c)
        curve_list.append(L_8c_3)
    point_ref = (Z3 + Z4 + Z5 + Z6 + Z8b + Z8c) / 6

    # Defining type of magnetization of the magnet
    if self.magnet_0:
        if self.magnet_0.type_magnetization == 0:
            type_mag = "_Radial"
        else:
            type_mag = "_Parallel"
    else:
        type_mag = "None"
    magnet_label = "HoleMagnet" + st + type_mag + "_N_R0_T0_S0"
    S2 = SurfLine(line_list=curve_list, label=magnet_label, point_ref=point_ref)

    # Air surface with magnet_0 and W1 > 0
    curve_list = list()
    curve_list.append(L_6_7)
    curve_list.append(L_7_8)
    if self.W2 > 0:  # if W2=0 Z8 = Z8b
        curve_list.append(L_8_8b)
    curve_list.append(L_8b_6)
    point_ref = (Z6 + Z7 + Z8 + Z8b) / 4

    S3 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Symmetry Air surface (W3) with magnet_1
    curve_list = list()
    curve_list.append(L_1s_2s)
    curve_list.append(L_2s_3s)
    curve_list.append(L_3s_8cs)
    curve_list.append(L_8cs_9s)
    if self.H4 > 0:
        curve_list.append(L_9s_10s)
    curve_list.append(L_10s_11s)
    if self.H4 > 0:
        curve_list.append(L_11s_1s)
    point_ref = (Z1s + Z2s + Z3s + Z8cs + Z9s + Z10s + Z11s) / 7

    S4 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # magnet_1 surface
    curve_list = list()
    if is_simplified:
        curve_list.append(L_8cs_3s.copy())
        curve_list.append(L_6s_8bs.copy())
    else:
        if Z3s != Z4s:  # Z3 == Z3 if H2 = 0
            curve_list.append(L_3s_4s.copy())
        curve_list.append(L_4s_5s.copy())
        if Z5s != Z6s:  # Z5 == Z6 if H2 = 0
            curve_list.append(L_5s_6s.copy())
        curve_list.append(L_6s_8bs.copy())
        curve_list.append(L_8bs_8cs.copy())
        curve_list.append(L_8cs_3s.copy())
    point_ref = (Z3s + Z4s + Z5s + Z6s + Z8bs + Z8cs) / 6
    # Defining type of magnetization of the magnet
    if self.magnet_1:
        if self.magnet_1.type_magnetization == 0:
            type_mag = "_Radial"
        else:
            type_mag = "_Parallel"
    else:
        type_mag = "None"
    magnet_label = "HoleMagnet" + st + type_mag + "_N_R0_T1_S0"
    S5 = SurfLine(line_list=curve_list, label=magnet_label, point_ref=point_ref)

    # Air surface with magnet_1 and W1 > 0
    curve_list = list()
    curve_list.append(L_6s_7s)
    curve_list.append(L_7s_8s)  # rad. bridge
    if self.W2 > 0:  # if W2=0: Z8s = Z8bs
        curve_list.append(L_8s_8bs)
    curve_list.append(L_8bs_6s)
    point_ref = (Z6s + Z7s + Z8s + Z8bs) / 4

    S6 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Air surface both magnets and W1 = 0 (S6 + S3)
    curve_list = list()
    curve_list.append(L_6_7.copy())
    curve_list.append(L_7_6s.copy())
    curve_list.append(L_6s_8bs.copy())
    if self.W2 > 0:  # If W2 = 0: Z8b = Z8 = Z8bs
        curve_list.append(L_8bs_8s.copy())
        curve_list.append(L_8s_8b.copy())
    curve_list.append(L_8b_6.copy())  # == Magnet_0_Left
    point_ref = (Z6 + Z7 + Z6s + Z8s + Z8bs + Z8b) / 6
    S7 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Air surface without magnet_0 and W1 > 0 (S1 + S2 + S3)
    curve_list = list()
    curve_list.append(L_1_2)
    curve_list.append(L_2_3)
    if self.H2 > 0:
        curve_list.append(L_3_4)
    curve_list.append(L_4_5)
    if self.H2 > 0:
        curve_list.append(L_5_6)
    curve_list.append(L_6_7)
    curve_list.append(L_7_8)
    curve_list.append(L_8_9)
    if self.H4 > 0:
        curve_list.append(L_9_10)
    curve_list.append(L_10_11)
    if self.H4 > 0:
        curve_list.append(L_11_1)
    point_ref = (Z1 + Z2 + Z3 + Z8c + Z9 + Z10 + Z11) / 7
    S8 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Air surface without magnet_1 and W1 > 0
    curve_list = list()
    curve_list.append(L_1s_2s)
    curve_list.append(L_2s_3s)
    if self.H2 > 0:
        curve_list.append(L_3s_4s)
    curve_list.append(L_4s_5s)
    if self.H2 > 0:
        curve_list.append(L_5s_6s)
    curve_list.append(L_6s_7s)
    curve_list.append(L_7s_8s)
    curve_list.append(L_8s_9s)
    if self.H4 > 0:
        curve_list.append(L_9s_10s)
    curve_list.append(
        Arc1(Z10s, Z11s, Rext - self.H1, is_trigo_direction=True, label="")
    )
    if self.H4 > 0:
        curve_list.append(L_11s_1s)
    point_ref = (Z1s + Z2s + Z3s + Z8cs + Z9s + Z10s + Z11s) / 7
    S9 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Air surface with magnet_0 without magnet_1 and W1 = 0
    # (S4 + S5 + S7)
    curve_list = list()
    curve_list.append(L_1s_2s.copy())
    curve_list.append(L_2s_3s.copy())
    if self.H2 > 0:
        curve_list.append(L_3s_4s.copy())
    curve_list.append(L_4s_5s.copy())
    if self.H2 > 0:
        curve_list.append(L_5s_6s.copy())
    curve_list.append(L_6s_7s.copy())
    curve_list.append(L_7s_6.copy())
    curve_list.append(L_6_8b.copy())
    if Z8b != Z8s:
        curve_list.append(L_8b_8s.copy())
    curve_list.append(L_8s_9s.copy())
    if self.H4 > 0:
        curve_list.append(L_9s_10s.copy())
    curve_list.append(L_10s_11s.copy())
    if self.H4 > 0:
        curve_list.append(L_11s_1s.copy())
    point_ref = (Z1s + Z2s + Z3s + Z8cs + Z9s + Z10s + Z11s) / 7
    S10 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Air surface with magnet_1 without magnet_0 and W1 = 0
    # (S1 + S2 + S7)
    curve_list = list()
    curve_list.append(L_1_2)
    curve_list.append(L_2_3)
    if self.H2 > 0:
        curve_list.append(L_3_4)
    curve_list.append(L_4_5)
    if self.H2 > 0:
        curve_list.append(L_5_6)
    curve_list.append(L_6_7)
    curve_list.append(L_7_6s)
    curve_list.append(L_6s_8bs)
    if Z8bs != Z8:
        curve_list.append(L_8bs_8)
    curve_list.append(L_8_9)
    if self.H4 > 0:
        curve_list.append(L_9_10)
    curve_list.append(L_10_11)
    if self.H4 > 0:
        curve_list.append(L_11_1)
    point_ref = (Z1 + Z2 + Z3 + Z8c + Z9 + Z10 + Z11) / 7
    S11 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Air surface without magnets and W1 = 0
    # (S4 + S5 + S7 + S2 + S1)
    curve_list = list()
    curve_list.append(L_1_2)
    curve_list.append(L_2_3)
    if Z3 != Z4:  # if H2 = 0
        curve_list.append(L_3_4)
    curve_list.append(L_4_5)
    if Z5 != Z6:  # if H2 = 0
        curve_list.append(L_5_6)
    curve_list.append(L_6_7)
    curve_list.append(L_7_6s)
    if Z5s != Z6s:
        curve_list.append(L_6s_5s)
    curve_list.append(L_5s_4s)
    if Z3s != Z4s:
        curve_list.append(L_4s_3s)
    curve_list.append(L_3s_2s)
    curve_list.append(L_2s_1s)
    if self.H4 > 0:
        curve_list.append(L_1s_11s)
    curve_list.append(L_11s_10s)
    if self.H4 > 0:
        curve_list.append(L_10s_9s)
    curve_list.append(L_9s_8s)
    curve_list.append(L_8s_9)
    if self.H4 > 0:
        curve_list.append(L_9_10)
    curve_list.append(L_10_11)
    if self.H4 > 0:
        curve_list.append(L_11_1)

    point_ref = (Z6 + Z8b + Z7 + Z8 + Z6s + Z8bs) / 6
    S12 = SurfLine(line_list=curve_list, label="Hole" + st, point_ref=point_ref)

    # Create the surface list by selecting the correct ones
    if self.magnet_0 and self.magnet_1 and self.W1 > 0:
        S1.label = S1.label + "_R0_T0_S0"  # Hole
        S3.label = S3.label + "_R0_T1_S0"  # Hole
        S6.label = S6.label + "_R0_T2_S0"  # Hole
        S4.label = S4.label + "_R0_T3_S0"  # Hole
        surf_list = [S1, S2, S3, S6, S5, S4]
    elif self.magnet_0 and self.magnet_1 and self.W1 == 0:
        S1.label = S1.label + "_R0_T0_S0"  # Hole
        S7.label = S7.label + "_R0_T1_S0"  # Hole
        S4.label = S4.label + "_R0_T2_S0"  # Hole
        surf_list = [S1, S2, S7, S5, S4]
    elif self.magnet_0 and not self.magnet_1 and self.W1 > 0:
        S1.label = S1.label + "_R0_T0_S0"  # Hole
        S3.label = S3.label + "_R0_T1_S0"  # Hole
        S9.label = S9.label + "_R0_T2_S0"  # Hole
        surf_list = [S1, S2, S3, S9]
    elif self.magnet_0 and not self.magnet_1 and self.W1 == 0:
        S1.label = S1.label + "_R0_T0_S0"  # Hole
        S10.label = S10.label + "_R0_T1_S0"  # Hole
        surf_list = [S1, S2, S10]
    elif not self.magnet_0 and self.magnet_1 and self.W1 > 0:
        S8.label = S8.label + "_R0_T0_S0"  # Hole
        S6.label = S6.label + "_R0_T1_S0"  # Hole
        S4.label = S4.label + "_R0_T2_S0"  # Hole
        surf_list = [S8, S6, S5, S4]
    elif not self.magnet_0 and self.magnet_1 and self.W1 == 0:
        S11.label = S11.label + "_R0_T0_S0"  # Hole
        S4.label = S4.label + "_R0_T1_S0"  # Hole
        surf_list = [S11, S5, S4]
    elif not self.magnet_0 and not self.magnet_1 and self.W1 > 0:
        S8.label = S8.label + "_R0_T0_S0"  # Hole
        S9.label = S9.label + "_R0_T1_S0"  # Hole
        surf_list = [S8, S9]
    elif not self.magnet_0 and not self.magnet_1 and self.W1 == 0:
        S12.label = S12.label + "_R0_T0_S0"  # Hole
        surf_list = [S12]

    # Apply the transformations
    for surf in surf_list:
        surf.rotate(alpha)
        surf.translate(delta)

    return surf_list



from ....Methods import NotImplementedYetError


def comp_force_nodal(self, output):
    """Compute the nodal air-gap surface force based on Maxwell Tensor (MT).

    Parameters
    ----------
    self : ForceMT
        A ForceMT object

    output : Output
        an Output object (to update)
    """

    raise NotImplementedYetError("Nodal force not available yet for Maxwell tensor")
